#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from typing import Dict, List, Optional, Union, Tuple
from datetime import date, datetime
from hsfs.core.feature_monitoring_config_api import FeatureMonitoringConfigApi
from hsfs.core.feature_monitoring_result import FeatureMonitoringResult
from hsfs.core.feature_monitoring_result_api import FeatureMonitoringResultApi
from hsfs.core import feature_monitoring_config as fmc
from hsfs.core.feature_descriptive_statistics import FeatureDescriptiveStatistics
from hsfs.client.exceptions import FeatureStoreException
from hsfs import util
from hsfs.core.job_api import JobApi


class FeatureMonitoringResultEngine:
    """Logic and helper methods to deal with results from a feature monitoring job.

    Attributes:
        feature_store_id: int. Id of the respective Feature Store.
        feature_group_id: int. Id of the feature group, if monitoring a feature group.
        feature_view_name: str. Name of the feature view, if monitoring a feature view.
        feature_view_version: int. Version of the feature view, if monitoring a feature view.
    """

    def __init__(
        self,
        feature_store_id: int,
        feature_group_id: Optional[int] = None,
        feature_view_name: Optional[str] = None,
        feature_view_version: Optional[int] = None,
        **kwargs,
    ):
        if feature_group_id is None:
            assert feature_view_name is not None
            assert feature_view_version is not None

        self._feature_store_id = feature_store_id
        self._feature_group_id = feature_group_id
        self._feature_view_name = feature_view_name
        self._feature_view_version = feature_view_version

        self._feature_monitoring_result_api = FeatureMonitoringResultApi(
            feature_store_id=feature_store_id,
            feature_group_id=feature_group_id,
            feature_view_name=feature_view_name,
            feature_view_version=feature_view_version,
        )
        self._feature_monitoring_config_api = FeatureMonitoringConfigApi(
            feature_store_id=feature_store_id,
            feature_group_id=feature_group_id,
            feature_view_name=feature_view_name,
            feature_view_version=feature_view_version,
        )
        self._job_api = JobApi()

    def build_feature_monitoring_result(
        self,
        config_id: int,
        feature_name: str,
        detection_statistics: Optional[FeatureDescriptiveStatistics] = None,
        reference_statistics: Optional[FeatureDescriptiveStatistics] = None,
        specific_value: Optional[Union[int, float]] = None,
        shift_detected: bool = False,
        difference: Optional[float] = None,
        raised_exception: bool = False,
        execution_id: Optional[int] = None,
        job_name: Optional[str] = None,
    ) -> FeatureMonitoringResult:
        """Build feature monitoring result.

        Args:
            config_id: int. Id of the feature monitoring configuration.
            execution_id: int. Id of the job execution.
            detection_statistics: FeatureDescriptiveStatistics. Statistics computed from the detection data.
            reference_statistics: Optional[Union[FeatureDescriptiveStatistics, int, float]]. Statistics computed from the reference data.
                Defaults to None if no reference is provided.
            shift_detected: bool. Whether a shift is detected between the detection and reference window.
                It is used to decide whether to trigger an alert.
            difference: Optional[float]. Difference between detection statistics and reference statistics.

        Returns:
            FeatureMonitoringResult. Saved Feature monitoring result.
        """
        monitoring_time = round(
            util.convert_event_time_to_timestamp(datetime.now()), -3
        )
        if execution_id is None and job_name is not None:
            execution_id = self.get_monitoring_job_execution_id(job_name)
        else:
            execution_id = 0
        detection_statistics_id = (
            detection_statistics.id
            if isinstance(detection_statistics, FeatureDescriptiveStatistics)
            else None
        )
        reference_statistics_id = (
            reference_statistics.id
            if isinstance(reference_statistics, FeatureDescriptiveStatistics)
            else None
        )
        if raised_exception and feature_name is None:
            # if feature name is null it is a whole entity monitoring job
            feature_name = ""

        return FeatureMonitoringResult(
            feature_store_id=self._feature_store_id,
            config_id=config_id,
            execution_id=execution_id,
            feature_name=feature_name,
            detection_statistics_id=detection_statistics_id,
            reference_statistics_id=reference_statistics_id,
            specific_value=specific_value,
            difference=difference,
            shift_detected=shift_detected,
            monitoring_time=monitoring_time,
            raised_exception=raised_exception,
            empty_detection_window=self.is_monitoring_window_empty(
                detection_statistics
            ),
            empty_reference_window=self.is_monitoring_window_empty(
                reference_statistics
            ),
        )

    def save_feature_monitoring_result(
        self,
        result: FeatureMonitoringResult,
    ) -> FeatureMonitoringResult:
        """Save feature monitoring result.

        Args:
            config_id: int. Id of the feature monitoring configuration.
            execution_id: int. Id of the job execution.
            detection_statistics: FeatureDescriptiveStatistics. Statistics computed from the detection data.
            reference_statistics: Optional[Union[FeatureDescriptiveStatistics, int, float]]. Statistics computed from the reference data.
                Defaults to None if no reference is provided.
            shift_detected: bool. Whether a shift is detected between the detection and reference window.
                It is used to decide whether to trigger an alert.
            difference: Optional[float]. Difference between detection statistics and reference statistics.

        Returns:
            FeatureMonitoringResult. Saved Feature monitoring result.
        """
        return self._feature_monitoring_result_api.create(
            result,
        )

    def fetch_all_feature_monitoring_results_by_config_id(
        self,
        config_id: int,
        start_time: Union[str, int, datetime, date, None] = None,
        end_time: Union[str, int, datetime, date, None] = None,
        with_statistics: bool = False,
    ) -> List[FeatureMonitoringResult]:
        """Fetch all feature monitoring results by config id.

        Args:
            config_id: int. Id of the feature monitoring configuration.
            start_time: Union[str, int, datetime, date, None].
                Query results with monitoring time greater than or equal to start_time.
            end_time: Union[str, int, datetime, date, None].
                Query results with monitoring time less than or equal to end_time.
            with_statistics: bool.
                Whether to include the statistics attached to the results or not

        Returns:
            List[FeatureMonitoringResult]. List of feature monitoring results.
        """

        query_params = self._build_query_params(
            start_time=start_time,
            end_time=end_time,
            with_statistics=with_statistics,
        )

        return self._feature_monitoring_result_api.get_by_config_id(
            config_id=config_id,
            query_params=query_params,
        )

    def get_feature_monitoring_results(
        self,
        config_id: Optional[int] = None,
        config_name: Optional[str] = None,
        start_time: Optional[Union[str, int, datetime, date]] = None,
        end_time: Optional[Union[str, int, datetime, date]] = None,
        with_statistics: bool = True,
    ) -> List[FeatureMonitoringResult]:
        """Convenience method to fetch feature monitoring results from an entity.

        Args:
            config_id: Optional[int]. Id of the feature monitoring configuration.
                Defaults to None if config_name is provided.
            config_name: Optional[str]. Name of the feature monitoring configuration.
                Defaults to None if config_id is provided.
            start_time: Optional[Union[str, int, datetime, date]].
                Query results with monitoring time greater than or equal to start_time.
            end_time: Optional[Union[str, int, datetime, date]].
                Query results with monitoring time less than or equal to end_time.
            with_statistics: bool. Whether to include the statistics attached to the results.
                Defaults to True. Set to False to fetch only monitoring metadata.

        Returns:
            List[FeatureMonitoringResult]. List of feature monitoring results.
        """
        if all([config_id is None, config_name is None]):
            raise ValueError(
                "Either config_id or config_name must be provided to fetch feature monitoring results."
            )
        elif all([config_id is not None, config_name is not None]):
            raise ValueError(
                "Only one of config_id or config_name can be provided to fetch feature monitoring results."
            )
        elif config_name is not None and isinstance(config_name, str):
            config = self._feature_monitoring_config_api.get_by_name(config_name)
            if not isinstance(config, fmc.FeatureMonitoringConfig):
                raise ValueError(
                    f"Feature monitoring configuration with name {config_name} does not exist."
                )
            config_id = config._id
        elif config_name is not None:
            raise TypeError(
                f"config_name must be of type str. Got {type(config_name)}."
            )
        elif config_id is not None and not isinstance(config_id, int):
            raise TypeError(f"config_id must be of type int. Got {type(config_id)}.")

        return self.fetch_all_feature_monitoring_results_by_config_id(
            config_id=config_id,
            start_time=start_time,
            end_time=end_time,
            with_statistics=with_statistics,
        )

    def _build_query_params(
        self,
        start_time: Union[str, int, datetime, date, None],
        end_time: Union[str, int, datetime, date, None],
        with_statistics: bool,
    ) -> Dict[str, Union[str, List[str]]]:
        """Build query parameters for feature monitoring result API calls.

        Args:
            start_time: Union[str, int, datetime, date, None].
                Query results with monitoring time greater than or equal to start_time.
            end_time: Union[str, int, datetime, date, None].
                Query results with monitoring time less than or equal to end_time.
            with_statistics: bool.
                Whether to include the statistics attached to the results or not

        Returns:
            Dict[str, Union[str, List[str]]]. Query parameters.
        """

        query_params = {"sort_by": "monitoring_time:desc"}

        filter_by = []
        if start_time:
            timestamp_start_time = util.convert_event_time_to_timestamp(start_time)
            filter_by.append(f"monitoring_time_gte:{timestamp_start_time}")
        if end_time:
            timestamp_end_time = util.convert_event_time_to_timestamp(end_time)
            filter_by.append(f"monitoring_time_lte:{timestamp_end_time}")
        if len(filter_by) > 0:
            query_params["filter_by"] = filter_by

        if with_statistics:
            query_params["expand"] = "statistics"

        return query_params

    def run_and_save_statistics_comparison(
        self,
        fm_config: "fmc.FeatureMonitoringConfig",
        detection_statistics: List[FeatureDescriptiveStatistics],
        reference_statistics: Optional[List[FeatureDescriptiveStatistics]] = None,
        specific_value: Optional[Union[int, float]] = None,
    ) -> List[FeatureMonitoringResult]:
        """Run and upload statistics comparison between detection and reference stats.

        Args:
            fm_config: FeatureMonitoringConfig. Feature monitoring configuration.
            detection_statistics: List[FeatureDescriptiveStatistics]. Computed statistics from detection data.
            reference_statistics: Optional[List[FeatureDescriptiveStatistics]]]. Computed statistics from reference data.
            specific_value: Optional[Union[int, float]]. A specific value to use as reference.

        Returns:
            Union[FeatureMonitoringResult, List[FeatureMonitoringResult]]. Feature monitoring result
        """
        if specific_value:
            # compare detection statistics against a specific value
            assert reference_statistics is None
            return self._run_and_save_statistics_comparison_specific_value(
                fm_config, detection_statistics, specific_value
            )

        if reference_statistics is not None:
            # compare detection statistics against reference statistics
            return self._run_and_save_statistics_comparison_reference_stats(
                fm_config, detection_statistics, reference_statistics
            )

        # otherwise, no comparison needed
        return [
            self.save_feature_monitoring_result(
                result=self.build_feature_monitoring_result(
                    config_id=fm_config.id,
                    feature_name=det_fds.feature_name,
                    detection_statistics=det_fds,
                )
            )
            for det_fds in detection_statistics
        ]

    def _run_and_save_statistics_comparison_specific_value(
        self,
        fm_config: "fmc.FeatureMonitoringConfig",
        detection_statistics: List[FeatureDescriptiveStatistics],
        specific_value: Union[int, float],
    ):
        assert len(detection_statistics) == 1, (
            "detection_statistics must a be a single FeatureDescriptiveStatistics object "
            "if reference_statistics or specific_value is provided."
        )
        det_fds = detection_statistics[0]
        difference, shift_detected = self.compute_difference_and_shift(
            fm_config=fm_config,
            detection_statistics=det_fds,
            specific_value=specific_value,
        )
        return [
            self.save_feature_monitoring_result(
                result=self.build_feature_monitoring_result(
                    config_id=fm_config.id,
                    feature_name=det_fds.feature_name,
                    detection_statistics=det_fds,
                    difference=difference,
                    shift_detected=shift_detected,
                    specific_value=specific_value,
                )
            )
        ]

    def _run_and_save_statistics_comparison_reference_stats(
        self,
        fm_config: "fmc.FeatureMonitoringConfig",
        detection_statistics: List[FeatureDescriptiveStatistics],
        reference_statistics: List[FeatureDescriptiveStatistics],
    ):
        assert len(reference_statistics) == len(
            detection_statistics
        ), "detection_statistics and reference_statistics must contain the same number of feature statistics"
        det_stats_set, ref_stats_set = set(), set()
        for det_fds, ref_fds in zip(detection_statistics, reference_statistics):
            det_stats_set.add((det_fds.feature_name, det_fds.feature_type))
            ref_stats_set.add((ref_fds.feature_name, ref_fds.feature_type))

        # check if detection and reference statistics contain the same features and feature types.
        # statistics computed on empty data will have feature type None, which can falsify the equality.
        # we shouldn't raise an exception in that case, so we just ignore it by checking the count statistic.
        if (
            (det_stats_set != ref_stats_set)
            and (detection_statistics[0].count != 0)
            and (reference_statistics[0].count != 0)
        ):
            raise FeatureStoreException(
                "Detection and reference statistics do not contain the same features or feature types"
            )

        # sort by feature name
        sorted_det_stats, sorted_ref_stats = (
            sorted(detection_statistics, key=lambda fds: fds.feature_name),
            sorted(reference_statistics, key=lambda fds: fds.feature_name),
        )

        fm_results = []
        for det_fds, ref_fds in zip(sorted_det_stats, sorted_ref_stats):
            difference, shift_detected = self.compute_difference_and_shift(
                fm_config=fm_config,
                detection_statistics=det_fds,
                reference_statistics=ref_fds,
            )
            fm_results.append(
                self.save_feature_monitoring_result(
                    result=self.build_feature_monitoring_result(
                        config_id=fm_config.id,
                        feature_name=det_fds.feature_name,
                        detection_statistics=det_fds,
                        reference_statistics=ref_fds,
                        difference=difference,
                        shift_detected=shift_detected,
                    )
                )
            )
        return fm_results

    def compute_difference_and_shift(
        self,
        fm_config: "fmc.FeatureMonitoringConfig",
        detection_statistics: FeatureDescriptiveStatistics,
        reference_statistics: Optional[FeatureDescriptiveStatistics] = None,
        specific_value: Optional[Union[int, float]] = None,
    ) -> Tuple[Optional[float], bool]:
        """Compute the difference and detect shift between the reference and detection statistics.

        Args:
            fm_config: FeatureMonitoringConfig. Feature monitoring configuration.
            detection_statistics: FeatureDescriptiveStatistics. Computed statistics from detection data.
            reference_statistics: Union[FeatureDescriptiveStatistics, int, float].
                Computed statistics from reference data, or a specific value to use as reference.

        Returns:
            `(float, bool)`. The difference between the reference and detection statistics,
                             and whether shift was detected or not
        """
        difference = self.compute_difference_between_stats(
            detection_statistics=detection_statistics,
            reference_statistics=reference_statistics,
            metric=fm_config.statistics_comparison_config["metric"].lower(),
            relative=fm_config.statistics_comparison_config["relative"],
            specific_value=specific_value,
        )
        if difference is None:
            return None, False  # no difference, no shift detected

        if fm_config.statistics_comparison_config["strict"]:
            shift_detected = (
                True
                if difference >= fm_config.statistics_comparison_config["threshold"]
                else False
            )
        else:
            shift_detected = (
                True
                if difference > fm_config.statistics_comparison_config["threshold"]
                else False
            )
        return difference, shift_detected

    def compute_difference_between_stats(
        self,
        detection_statistics: FeatureDescriptiveStatistics,
        metric: str,
        relative: bool = False,
        reference_statistics: Optional[FeatureDescriptiveStatistics] = None,
        specific_value: Optional[Union[int, float]] = None,
    ) -> Optional[float]:
        """Compute the difference between the reference and detection statistics.

        Args:
            detection_statistics: `FeatureDescriptiveStatistics`. Computed statistics from detection data.
            reference_statistics: `Optional[FeatureDescriptiveStatistics]`.
                Computed statistics from reference data, or a specific value to use as reference
            metric: `str`. The metric to compute the difference for.
            relative: `bool`. Whether to compute the relative difference or not.
            specific_value: `Optional[Union[int, float]]`. A specific value to use as reference.

        Returns:
            `Optional[float]`. The difference between the reference and detection statistics, or None if there
                               are no values to compare
        """
        if (detection_statistics.count == 0) or (
            specific_value is None and reference_statistics.count == 0
        ):
            # if stats on empty data, no difference can be computed
            return None

        # otherwise, both detection and reference value can be obtained
        detection_value = detection_statistics.get_value(metric)
        reference_value = (
            specific_value
            if specific_value is not None
            else reference_statistics.get_value(metric)
        )
        return self.compute_difference_between_specific_values(
            detection_value, reference_value, relative
        )

    def compute_difference_between_specific_values(
        self,
        detection_value: Union[int, float],
        reference_value: Union[int, float],
        relative: bool = False,
    ) -> float:
        """Compute the difference between a reference and detection value.

        Args:
            detection_value: Union[int, float]. The detection value.
            reference_value: Union[int, float]. The reference value
            relative: `bool`. Whether to compute the relative difference or not.

        Returns:
            `float`. The difference between the reference and detection values.
        """

        diff = abs(detection_value - reference_value)
        if relative:
            if reference_value == 0:
                return float("inf")
            else:
                return diff / reference_value
        else:
            return diff

    def get_monitoring_job_execution_id(
        self,
        job_name: str,
    ) -> int:
        """Get the execution id of the last execution of the monitoring job.

        The last execution is assumed to be the current execution.
        The id defaults to 0 if no execution is found.

        Args:
            job_name: str. Name of the monitoring job.

        Returns:
            int. Id of the last execution of the monitoring job.
                It is assumed to be the current execution.
        """
        execution = self._job_api.last_execution(self._job_api.get(name=job_name))
        return (
            execution[0]._id
            if isinstance(execution, list) and len(execution) > 0
            else 0
        )

    def is_monitoring_window_empty(
        self,
        monitoring_window_statistics: Optional[FeatureDescriptiveStatistics] = None,
    ) -> bool:
        """Check if the monitoring window is empty.

        Args:
            monitoring_window_statistics: FeatureDescriptiveStatistics. Statistics computed for the monitoring window.

        Returns:
            bool. Whether the monitoring window is empty or not.
        """
        if monitoring_window_statistics is None:
            return True
        elif monitoring_window_statistics.count == 0:
            return True
        else:
            return False

    def save_feature_monitoring_result_with_exception(
        self,
        config_id: int,
        job_name: str,
        feature_name: Optional[str] = None,
    ) -> "FeatureMonitoringResult":
        """Save feature monitoring result with raised_exception flag.

        Args:
            config_id: int. Id of the feature monitoring configuration.
            job_name: str. Name of the monitoring job.

        Returns:
            FeatureMonitoringResult. Saved Feature monitoring result.
        """
        return self.save_feature_monitoring_result(
            result=self.build_feature_monitoring_result(
                config_id=config_id,
                job_name=job_name,
                raised_exception=True,
                feature_name=feature_name,
            ),
        )



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import json
import logging
from typing import Any, Dict, List, Optional
from warnings import warn

import requests
from furl import furl
from hsfs import client
from hsfs.client.exceptions import FeatureStoreException
from hsfs.core import variable_api


_logger = logging.getLogger(__name__)

_online_store_rest_client = None


def init_or_reset_online_store_rest_client(
    optional_config: Optional[Dict[str, Any]] = None, reset_client: bool = False
):
    global _online_store_rest_client
    if not _online_store_rest_client:
        _online_store_rest_client = OnlineStoreRestClientSingleton(
            optional_config=optional_config
        )
    elif reset_client:
        _online_store_rest_client.reset_client(optional_config=optional_config)
    else:
        warn(
            "Online Store Rest Client is already initialised. To reset connection or/and override configuration, "
            + "use reset_online_store_rest_client flag.",
            stacklevel=2,
        )


def get_instance() -> "OnlineStoreRestClientSingleton":
    global _online_store_rest_client
    if not _online_store_rest_client:
        _logger.warning(
            "Online Store Rest Client is not initialised. Initialising with default configuration."
        )
        _online_store_rest_client = OnlineStoreRestClientSingleton()
    _logger.debug("Accessing global Online Store Rest Client instance.")
    return _online_store_rest_client


_DEFAULT_ONLINE_STORE_REST_CLIENT_PORT = 4406
_DEFAULT_ONLINE_STORE_REST_CLIENT_TIMEOUT_MS = 600
_DEFAULT_ONLINE_STORE_REST_CLIENT_VERIFY_CERTS = True
_DEFAULT_ONLINE_STORE_REST_CLIENT_USE_SSL = True
_DEFAULT_ONLINE_STORE_REST_CLIENT_SERVER_API_VERSION = "0.1.0"
_DEFAULT_ONLINE_STORE_REST_CLIENT_HTTP_AUTHORIZATION = "X-API-KEY"


class OnlineStoreRestClientSingleton:
    HOST = "host"
    PORT = "port"
    VERIFY_CERTS = "verify_certs"
    USE_SSL = "use_ssl"
    CA_CERTS = "ca_certs"
    HTTP_AUTHORIZATION = "http_authorization"
    TIMEOUT = "timeout"
    SERVER_API_VERSION = "server_api_version"
    API_KEY = "api_key"

    def __init__(self, optional_config: Optional[Dict[str, Any]] = None):
        _logger.info(
            f"Initialising Online Store Rest Client {'with optional configuration' if optional_config else ''}."
        )
        if optional_config:
            _logger.debug(f"Optional Config: {optional_config}")
        self._check_hopsworks_connection()
        self.variable_api = variable_api.VariableApi()
        self._auth: client.auth.OnlineStoreKeyAuth
        self._session: requests.Session
        self._current_config: Dict[str, Any]
        self._base_url: furl
        self._setup_rest_client(
            optional_config=optional_config, use_current_config=False
        )
        self.is_connected()

    def reset_client(self, optional_config: Optional[Dict[str, Any]] = None):
        _logger.info(
            f"Resetting Online Store Rest Client {'with optional configuration' if optional_config else ''}."
        )
        if optional_config:
            _logger.debug(f"Optional Config: {optional_config}")
        self._check_hopsworks_connection()
        if hasattr(self, "_session") and self._session:
            _logger.debug("Closing existing session.")
            self._session.close()
            delattr(self, "_session")
        self._setup_rest_client(
            optional_config=optional_config,
            use_current_config=False if optional_config else True,
        )

    def _setup_rest_client(
        self,
        optional_config: Optional[Dict[str, Any]] = None,
        use_current_config: bool = True,
    ):
        _logger.info("Setting up Online Store Rest Client.")
        if optional_config and not isinstance(optional_config, dict):
            raise ValueError(
                "optional_config must be a dictionary. See documentation for allowed keys and values."
            )
        _logger.debug(f"Optional Config: {json.dumps(optional_config, indent=2)}")
        if not use_current_config:
            _logger.debug(
                "Retrieving default configuration for Online Store REST Client."
            )
            self._current_config = self._get_default_client_config()
        if optional_config:
            _logger.debug(
                "Updating default configuration with provided optional configuration."
            )
            self._current_config.update(optional_config)

        self._set_auth(optional_config)
        if not hasattr(self, "_session") or not self._session:
            _logger.debug("Initialising new requests session.")
            self._session = requests.Session()
        else:
            raise ValueError(
                "Use the init_or_reset_online_store_connection method with reset_connection flag set "
                + "to True to reset the online_store_client_connection"
            )
        if not self._current_config[self.VERIFY_CERTS]:
            _logger.warning(
                "Disabling SSL certificate verification. This is not recommended for production environments."
            )
            self._session.verify = False
        else:
            _logger.debug(
                f"Setting SSL certificate verification using CA Certs path: {self._current_config[self.CA_CERTS]}"
            )
            self._session.verify = self._current_config[self.CA_CERTS]

        # Set base_url
        scheme = "https" if self._current_config[self.USE_SSL] else "http"
        self._base_url = furl(
            f"{scheme}://{self._current_config[self.HOST]}:{self._current_config[self.PORT]}/{self._current_config[self.SERVER_API_VERSION]}"
        )

        assert (
            self._session is not None
        ), "Online Store REST Client failed to initialise."
        assert (
            self._auth is not None
        ), "Online Store REST Client Authentication failed to initialise. Check API Key."
        assert (
            self._base_url is not None
        ), "Online Store REST Client Base URL failed to initialise. Check host and port parameters."
        assert (
            self._current_config is not None
        ), "Online Store REST Client Configuration failed to initialise."

    def _get_default_client_config(self) -> Dict[str, Any]:
        _logger.debug("Retrieving default configuration for Online Store REST Client.")
        default_config = self._get_default_static_parameters_config()
        default_config.update(self._get_default_dynamic_parameters_config())
        return default_config

    def _get_default_static_parameters_config(self) -> Dict[str, Any]:
        _logger.debug(
            "Retrieving default static configuration for Online Store REST Client."
        )
        return {
            self.TIMEOUT: _DEFAULT_ONLINE_STORE_REST_CLIENT_TIMEOUT_MS,
            self.VERIFY_CERTS: _DEFAULT_ONLINE_STORE_REST_CLIENT_VERIFY_CERTS,
            self.USE_SSL: _DEFAULT_ONLINE_STORE_REST_CLIENT_USE_SSL,
            self.SERVER_API_VERSION: _DEFAULT_ONLINE_STORE_REST_CLIENT_SERVER_API_VERSION,
            self.HTTP_AUTHORIZATION: _DEFAULT_ONLINE_STORE_REST_CLIENT_HTTP_AUTHORIZATION,
        }

    def _get_default_dynamic_parameters_config(
        self,
    ) -> Dict[str, Any]:
        _logger.debug(
            "Retrieving default dynamic configuration for Online Store REST Client."
        )
        url = furl(self._get_rondb_rest_server_endpoint())
        _logger.debug(f"Default RonDB Rest Server host and port: {url.host}:{url.port}")
        _logger.debug(
            f"Using CA Certs from Hopsworks Client: {client.get_instance()._get_ca_chain_path()}"
        )
        return {
            self.HOST: url.host,
            self.PORT: url.port,
            self.CA_CERTS: client.get_instance()._get_ca_chain_path(),
        }

    def _get_rondb_rest_server_endpoint(self) -> str:
        """Retrieve RonDB Rest Server endpoint based on whether the client is running internally or externally.
        If the client is running externally, the endpoint is retrieved via the loadbalancer.
        If the client is running internally, the endpoint is retrieved via (consul) service discovery.
        The default port for the RonDB Rest Server is 4406 and always used unless specifying a different port
        in the configuration.
        Returns:
            str: RonDB Rest Server endpoint with default port.
        """
        if client.get_instance()._is_external():
            _logger.debug(
                "External Online Store REST Client : Retrieving RonDB Rest Server endpoint via loadbalancer."
            )
            external_domain = self.variable_api.get_loadbalancer_external_domain()
            if external_domain == "":
                _logger.debug(
                    "External Online Store REST Client : Loadbalancer external domain is not set. Using client host as endpoint."
                )
                external_domain = client.get_instance().host
            default_url = (
                f"https://{external_domain}:{_DEFAULT_ONLINE_STORE_REST_CLIENT_PORT}"
            )
            _logger.debug(
                f"External Online Store REST Client : Default RonDB Rest Server endpoint: {default_url}"
            )
            return default_url
        else:
            _logger.debug(
                "Internal Online Store REST Client : Retrieving RonDB Rest Server endpoint via service discovery."
            )
            service_discovery_domain = self.variable_api.get_service_discovery_domain()
            if service_discovery_domain == "":
                raise FeatureStoreException("Service discovery domain is not set.")
            default_url = f"https://rdrs.service.{service_discovery_domain}:{_DEFAULT_ONLINE_STORE_REST_CLIENT_PORT}"
            _logger.debug(
                f"Internal Online Store REST Client : Default RonDB Rest Server endpoint: {default_url}"
            )
            return default_url

    def send_request(
        self,
        method: str,
        path_params: List[str],
        headers: Optional[Dict[str, Any]] = None,
        data: Optional[str] = None,
    ) -> requests.Response:
        url = self._base_url.copy()
        url.path.segments.extend(path_params)
        _logger.info(f"Sending {method} request to {url.url}.")
        _logger.debug(f"Provided Data: {data}")
        _logger.debug(f"Provided Headers: {headers}")
        prepped_request = self._session.prepare_request(
            requests.Request(
                method, url=url.url, headers=headers, data=data, auth=self.auth
            )
        )
        return self._session.send(
            prepped_request,
            timeout=self._current_config[self.TIMEOUT] / 1000,
        )

    def _check_hopsworks_connection(self) -> None:
        _logger.debug("Checking Hopsworks connection.")
        assert (
            client.get_instance() is not None and client.get_instance()._connected
        ), """Hopsworks Client is not connected. Please connect to Hopsworks cluster
            via hopsworks.login or hsfs.connection before initialising the Online Store REST Client.
            """
        _logger.debug("Hopsworks connection is active.")

    def _set_auth(self, optional_config: Optional[Dict[str, Any]] = None) -> None:
        """Set authentication object for the Online Store REST Client.
        RonDB Rest Server uses Hopsworks Api Key to authenticate requests via the X-API-KEY header by default.
        The api key determines the permissions of the user making the request for access to a given Feature Store.
        """
        _logger.debug("Setting authentication for Online Store REST Client.")
        if client.get_instance()._is_external():
            assert hasattr(
                client.get_instance()._auth, "_token"
            ), "External client must use API Key authentication. Contact your system administrator."
            _logger.debug(
                "External Online Store REST Client : Setting authentication using Hopsworks Client API Key."
            )
            self._auth = client.auth.OnlineStoreKeyAuth(
                client.get_instance()._auth._token
            )
        elif isinstance(optional_config, dict) and optional_config.get(
            self.API_KEY, False
        ):
            _logger.debug(
                "Setting authentication using provided API Key from optional configuration."
            )
            self._auth = client.auth.OnlineStoreKeyAuth(optional_config[self.API_KEY])
        elif hasattr(self, "_auth") and self._auth is not None:
            _logger.debug(
                "Authentication for Online Store REST Client is already set. Using existing authentication api key."
            )
        else:
            raise FeatureStoreException(
                "RonDB Rest Server uses Hopsworks Api Key to authenticate request."
                + f"Provide a configuration with the {self.API_KEY} key."
            )

    def is_connected(self):
        """If Online Store Rest Client is initialised, ping RonDB Rest Server to ensure connection is active."""
        if self._session is None:
            _logger.debug(
                "Checking Online Store REST Client is connected. Session is not initialised."
            )
            raise FeatureStoreException("Online Store REST Client is not initialised.")

        _logger.debug(
            "Checking Online Store REST Client is connected. Pinging RonDB Rest Server."
        )
        if not self.send_request("GET", ["ping"]):
            warn("Ping failed, RonDB Rest Server is not reachable.", stacklevel=2)
            return False
        return True

    @property
    def session(self) -> requests.Session:
        """Requests session object used to send requests to the Online Store REST API."""
        return self._session

    @property
    def base_url(self) -> furl:
        """Base URL for the Online Store REST API.
        This the url of the RonDB REST Server and should not be confused with the Opensearch Vector DB which also serves as an Online Store for features belonging to Feature Group containing embeddings."""
        return self._base_url

    @property
    def current_config(self) -> Dict[str, Any]:
        """Current configuration of the Online Store REST Client."""
        return self._current_config

    @property
    def auth(self) -> "client.auth.OnlineStoreKeyAuth":
        """Authentication object used to authenticate requests to the Online Store REST API.
        Extends the requests.auth.AuthBase class.
        """
        return self._auth



#
#   Copyright 2021 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json
import humps

from hsfs import util


class User:
    def __init__(
        self,
        username=None,
        email=None,
        first_name=None,
        last_name=None,
        status=None,
        secret=None,
        chosen_password=None,
        repeated_password=None,
        tos=None,
        two_factor=None,
        tours_state=None,
        max_num_projects=None,
        test_user=None,
        user_account_type=None,
        num_active_projects=None,
        num_remaining_projects=None,
        **kwargs,
    ):
        self._username = username
        self._email = email
        self._first_name = first_name
        self._last_name = last_name
        self._status = status
        self._secret = secret
        self._chosen_password = chosen_password
        self._repeated_password = repeated_password
        self._tos = tos
        self._two_factor = two_factor
        self._tours_state = tours_state
        self._max_num_projects = max_num_projects
        self._test_user = test_user
        self._user_account_type = user_account_type
        self._num_active_projects = num_active_projects
        self._num_remaining_projects = num_remaining_projects

    @classmethod
    def from_response_json(cls, json_dict):
        if json_dict:
            json_decamelized = humps.decamelize(json_dict)
            return cls(**json_decamelized)
        else:
            return None

    def json(self):
        return json.dumps(self, cls=util.FeatureStoreEncoder)

    @property
    def username(self):
        return self._username

    @property
    def email(self):
        return self._email

    @property
    def first_name(self):
        return self._first_name

    @property
    def last_name(self):
        return self._last_name

    @property
    def status(self):
        return self._status

    @property
    def secret(self):
        return self._secret

    @property
    def chosen_password(self):
        return self._chosen_password

    @property
    def repeated_password(self):
        return self._repeated_password

    @property
    def tos(self):
        return self._tos

    @property
    def two_factor(self):
        return self._two_factor

    @property
    def tours_state(self):
        return self._tours_state

    @property
    def max_num_projects(self):
        return self._max_num_projects

    @property
    def test_user(self):
        return self._test_user

    @property
    def user_account_type(self):
        return self._user_account_type

    @property
    def num_active_projects(self):
        return self._num_active_projects

    @property
    def num_remaining_projects(self):
        return self._num_remaining_projects



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json
import humps
from typing import List, Optional, Union
from hsfs.util import FeatureStoreEncoder
from enum import Enum

from hsfs.core import monitoring_window_config_engine


class WindowConfigType(str, Enum):
    ALL_TIME = "ALL_TIME"
    ROLLING_TIME = "ROLLING_TIME"
    TRAINING_DATASET = "TRAINING_DATASET"
    SPECIFIC_VALUE = "SPECIFIC_VALUE"

    @classmethod
    def list_str(cls) -> List[str]:
        return list(map(lambda c: c.value, cls))

    @classmethod
    def list(cls) -> List["WindowConfigType"]:
        return list(map(lambda c: c, cls))

    @classmethod
    def from_str(cls, value: str) -> "WindowConfigType":
        if value in cls.list_str():
            return cls(value)
        else:
            raise ValueError(
                f"Invalid value {value} for WindowConfigType, allowed values are {cls.list_str()}"
            )

    def __eq__(self, other):
        if isinstance(other, str):
            return self.value == other
        if isinstance(other, self.__class__):
            return self is other
        return False

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return self.value


class MonitoringWindowConfig:
    DEFAULT_ROW_PERCENTAGE = 1.0

    def __init__(
        self,
        id: Optional[int] = None,
        window_config_type: Optional[
            Union[str, WindowConfigType]
        ] = WindowConfigType.SPECIFIC_VALUE,
        time_offset: Optional[str] = None,
        window_length: Optional[str] = None,
        training_dataset_version: Optional[int] = None,
        specific_value: Optional[float] = None,
        row_percentage: Optional[float] = None,
        **kwargs,
    ):
        """Configuration to define the slice of data to compute statistics on.

        !!! example
            ```python3
            # Detection or reference window
            ## Rolling Time Window
            monitoring_window_config = MonitoringWindowConfig(
                time_offset="1d", # data inserted up to 1 day ago
                window_length="1h", # data inserted until an one hour after time_offset
                row_percentage=0.2, # include only 20% of the rows when computing statistics
            )

            ## Rolling Commit Window (not supported yet)
            monitoring_window_config = MonitoringWindowConfig(
                commit_offset=10, # data inserted up to 10 commits ago
                commit_num=5, # include 5 commits after commit_offset
            )

            # Only available for reference window
            ## Specific value
            monitoring_window_config = MonitoringWindowConfig(specific_value=0.5)

            ## Training dataset
            monitoring_window_config = MonitoringWindowConfig(
                training_dataset_version=my_training_dataset.version
            )
            ```

        # Arguments
            id: int, optional
                The id of the monitoring window config.
            window_config_type: str, optional
                The type of the monitoring window config. One of ROLLING_TIME,
                TRAINING_DATASET, SPECIFIC_VALUE.
            time_offset: str, optional
                The time offset of the monitoring window config. Only used for
                INSERT and SNAPSHOT window config types.
            window_length: str, optional
                The window length of the monitoring window config. Only used for
                INSERT and SNAPSHOT window config types.
            training_dataset_version: int, optional
                The version of the training dataset to use as reference. Only used for
                TRAINING_DATASET window config type.
            specific_value: float, optional
                The specific value to use as reference. Only used for SPECIFIC_VALUE
                window config type.
            row_percentage: float, optional
                The fraction of rows to use when computing statistics [0, 1.0]. Only used
                for ROLLING_TIME and ALL_TIME window config types.

        # Raises
            AttributeError: If window_config_type is not one of INSERT, SNAPSHOT,
                BATCH, TRAINING_DATASET, SPECIFIC_VALUE.
        """
        self._id = id
        self._window_config_type = None
        self.window_config_type = window_config_type
        self._time_offset = time_offset
        self._window_length = window_length
        self._training_dataset_version = training_dataset_version
        self._specific_value = specific_value

        if self.window_config_type in [
            WindowConfigType.SPECIFIC_VALUE,
            WindowConfigType.TRAINING_DATASET,
        ]:
            self._row_percentage = None
        else:
            self.row_percentage = row_percentage

        self._window_config_engine = (
            monitoring_window_config_engine.MonitoringWindowConfigEngine()
        )

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        return cls(**json_decamelized)

    def to_dict(self):
        the_dict = {
            "id": self._id,
            "windowConfigType": self._window_config_type,
        }

        if (
            self._window_config_type == WindowConfigType.ROLLING_TIME
            or self._window_config_type == WindowConfigType.ALL_TIME
        ):
            the_dict["timeOffset"] = self._time_offset
            the_dict["windowLength"] = self._window_length
            the_dict["rowPercentage"] = self.row_percentage
        elif self._window_config_type == WindowConfigType.SPECIFIC_VALUE:
            the_dict["specificValue"] = self._specific_value
        elif self._window_config_type == WindowConfigType.TRAINING_DATASET:
            the_dict["trainingDatasetVersion"] = self._training_dataset_version

        return the_dict

    def json(self) -> str:
        return json.dumps(self, cls=FeatureStoreEncoder)

    def __str__(self):
        return self.json()

    def __repr__(self):
        return f"MonitoringWindowConfig({self._window_config_type!r})"

    @property
    def id(self) -> Optional[int]:
        """Id of the window configuration."""
        return self._id

    @property
    def window_config_type(self) -> WindowConfigType:
        """Type of the window. It can be one of `ALL_TIME`, `ROLLING_TIME`, `TRAINING_DATASET` or `SPECIFIC_VALUE`."""
        return self._window_config_type

    @window_config_type.setter
    def window_config_type(self, window_config_type: Union[WindowConfigType, str]):
        if self._window_config_type is not None:
            raise AttributeError("window_config_type is a read-only attribute.")

        if isinstance(window_config_type, WindowConfigType):
            self._window_config_type = window_config_type
            return

        if not isinstance(window_config_type, str):
            raise TypeError(
                "window_config_type must be a string or WindowConfigType. "
                "Allowed value are" + str(WindowConfigType.list_str())
            )
        elif window_config_type not in WindowConfigType.list_str():
            raise ValueError(
                "window_config_type must be one of "
                + str(WindowConfigType.list_str())
                + "."
            )
        else:
            self._window_config_type = WindowConfigType.from_str(window_config_type)

    @property
    def time_offset(self) -> Optional[str]:
        """The time offset from the current time to the start of the time window. Only used for windows of type `ROLLING_TIME`."""
        return self._time_offset

    @property
    def window_length(self) -> Optional[str]:
        """The length of the time window. Only used for windows of type `ROLLING_TIME`."""
        return self._window_length

    @window_length.setter
    def window_length(self, window_length: Optional[str]):
        if window_length is None:
            self._window_length = None
        elif self._window_config_type != WindowConfigType.ROLLING_TIME:
            raise AttributeError(
                "Window length can only be set for if window_config_type is ROLLING_TIME."
            )
        elif isinstance(window_length, str):
            self._window_config_engine.time_range_str_to_time_delta(window_length)
            self._window_length = window_length
        else:
            raise TypeError("window_length must be a string.")

    @property
    def training_dataset_version(self) -> Optional[int]:
        """The version of the training dataset to use as reference. Only used for windows of type `TRAINING_DATASET`."""
        return self._training_dataset_version

    @training_dataset_version.setter
    def training_dataset_version(self, training_dataset_version: Optional[int]):
        if (
            self._window_config_type != WindowConfigType.TRAINING_DATASET
            and training_dataset_version is not None
        ):
            raise AttributeError(
                "training_dataset_version can only be set for if window_config_type is TRAINING_DATASET."
            )
        self._training_dataset_version = training_dataset_version

    @property
    def specific_value(self) -> Optional[float]:
        """The specific value to use as reference. Only used for windows of type `SPECIFIC_VALUE`."""
        return self._specific_value

    @specific_value.setter
    def specific_value(self, specific_value: Optional[float]):
        if (
            self._window_config_type != WindowConfigType.SPECIFIC_VALUE
            and specific_value is not None
        ):
            raise AttributeError(
                "Specific value can only be set for if window_config_type is SPECIFIC_VALUE."
            )
        self._specific_value = specific_value

    @property
    def row_percentage(self) -> Optional[float]:
        """The percentage of rows to fetch and compute the statistics on. Only used for windows of type `ROLLING_TIME` and `ALL_TIME`."""
        return self._row_percentage

    @row_percentage.setter
    def row_percentage(self, row_percentage: Optional[float]):
        if self.window_config_type in [
            WindowConfigType.SPECIFIC_VALUE,
            WindowConfigType.TRAINING_DATASET,
        ]:
            raise AttributeError(
                "Row percentage can only be set for ROLLING_TIME and ALL_TIME"
                " window config types."
            )

        if isinstance(row_percentage, int) or isinstance(row_percentage, float):
            row_percentage = float(row_percentage)
            if row_percentage <= 0.0 or row_percentage > 1.0:
                raise ValueError("Row percentage must be a float between 0 and 1.")
            self._row_percentage = row_percentage
        elif row_percentage is None:
            self._row_percentage = self.DEFAULT_ROW_PERCENTAGE
        else:
            raise TypeError("Row percentage must be a float between 0 and 1.")



#
#   Copyright 2023 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#


try:
    from pyspark.sql import SparkSession
except ImportError:
    pass

from hsfs.engine import spark


class Engine(spark.Engine):
    def __init__(self):
        self._spark_session = SparkSession.builder.getOrCreate()
        self._spark_context = self._spark_session.sparkContext
        self._jvm = self._spark_context._jvm

        super().__init__()

    def _sql_offline(self, sql_query, feature_store):
        # Spark no metastore does not require the
        return self._spark_session.sql(sql_query)



#
#   Copyright 2022 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from typing import Union, List, Dict
from hsfs.core import validation_result_api
from datetime import datetime, date
from hsfs.util import convert_event_time_to_timestamp

from hsfs.ge_validation_result import ValidationResult
from great_expectations.core import ExpectationValidationResult


class ValidationResultEngine:
    def __init__(self, feature_store_id: int, feature_group_id: int):
        """Validation Result engine.

        :param feature_store_id: id of the respective featurestore
        :type feature_store_id: int
        :param feature_group_id: id of the featuregroup it is attached to
        :type feature_group_id: int
        """
        self._validation_result_api = validation_result_api.ValidationResultApi(
            feature_store_id=feature_store_id, feature_group_id=feature_group_id
        )

    def get_validation_history(
        self,
        expectation_id: int,
        start_validation_time: Union[str, int, datetime, date, None] = None,
        end_validation_time: Union[str, int, datetime, date, None] = None,
        filter_by: List[str] = [],
        ge_type: bool = True,
    ) -> Union[List[ValidationResult], List[ExpectationValidationResult]]:
        """Get Validation Results relevant to an Expectation specified by expectation_id.

        :param expectation_id: id of the expectation for which to fetch the validation history
        :type expectation_id: int
        :param ingestion_only: retrieve only validation result linked to data ingested in the Feature Group
        :type ingestion_only: bool
        :param rejected_only: retrieve only validation result linked to data not ingested in the Feature Group
        :type rejected_only: bool
        :param ge_type: whether to convert Hopsworks object to native Great Expectations object
        :type ge_type: bool
        :param start_validation_time: retrieve validation result posterior to start_validation_time.
        Supported format include timestamps(int), datetime, date or string formatted to be datutils parsable.
        :type start_validation_time: Union[str, int, datetime, date, None]
        :param end_validation_time: retrieve validation result anterior to end_validation_time
        Supported format include timestamps(int), datetime, date or string formatted to be datutils parsable.
        :type end_validation_time: Union[str, int, datetime, date, None]
        """
        query_params = self._build_query_params(
            filter_by=filter_by,
            start_validation_time=start_validation_time,
            end_validation_time=end_validation_time,
        )

        history = self._validation_result_api.get_validation_history(
            expectation_id=expectation_id, query_params=query_params
        )

        if isinstance(history, ValidationResult):
            history = [history]

        if ge_type:
            return [result.to_ge_type() for result in history]
        else:
            return history

    def _build_query_params(
        self,
        filter_by: List[str] = [],
        start_validation_time: Union[str, int, datetime, date, None] = None,
        end_validation_time: Union[str, int, datetime, date, None] = None,
    ) -> Dict[str, str]:
        query_params = {"filter_by": [], "sort_by": "validation_time:desc"}
        allowed_ingestion_filters = [
            "INGESTED",
            "REJECTED",
            "UNKNOWN",
            "EXPERIMENT",
            "FG_DATA",
        ]
        ingestion_filters = []
        for ingestion_filter in filter_by:
            if ingestion_filter.upper() in allowed_ingestion_filters:
                ingestion_filters.append(
                    f"ingestion_result_eq:{ingestion_filter.upper()}"
                )
            else:
                raise ValueError(
                    f"Illegal Value {ingestion_filter} in filter_by."
                    + f"Allowed values are {', '.join(allowed_ingestion_filters)}"
                )

        query_params["filter_by"].extend(ingestion_filters)

        if start_validation_time and end_validation_time:
            if convert_event_time_to_timestamp(
                start_validation_time
            ) > convert_event_time_to_timestamp(end_validation_time):
                raise ValueError(
                    f"start_validation_time : {start_validation_time} is posterior to end_validation_time : {end_validation_time}"
                )

        if start_validation_time:
            query_params["filter_by"].append(
                "validation_time_gte:"
                + str(convert_event_time_to_timestamp(start_validation_time))
            )
        if end_validation_time:
            query_params["filter_by"].append(
                "validation_time_lte:"
                + str(convert_event_time_to_timestamp(end_validation_time))
            )

        return query_params



#
#   Copyright 2021 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from typing import List
from hsfs.client.exceptions import FeatureStoreException
from hsfs.core.feature_descriptive_statistics import FeatureDescriptiveStatistics


class BuiltInTransformationFunction:
    def __init__(self, method):
        self._method = method.lower()

    @staticmethod
    def min_max_scaler_stats(
        feature_descriptive_stats: List[FeatureDescriptiveStatistics], feature_name: str
    ):
        min_value = None
        max_value = None
        for stats in feature_descriptive_stats:
            if stats.feature_name == feature_name:
                if stats.feature_type not in ["Integral", "Fractional", "Decimal"]:
                    raise ValueError("Can't compute min_max_scaler for this type")
                min_value = stats.min
                max_value = stats.max

        if min_value is None or max_value is None:
            raise FeatureStoreException(
                "Feature {feature_name:} doesn't have minimum and/or maximum values computed. Thus can't use "
                "min_max_scaler method".format(feature_name=feature_name)
            )
        return min_value, max_value

    @staticmethod
    def standard_scaler_stats(
        feature_descriptive_stats: List[FeatureDescriptiveStatistics], feature_name: str
    ):
        mean = None
        std_dev = None
        for stats in feature_descriptive_stats:
            if stats.feature_name == feature_name:
                if stats.feature_type not in ["Integral", "Fractional", "Decimal"]:
                    raise ValueError("Can't compute standard_scaler for this type")
                mean = stats.mean
                std_dev = stats.stddev

        if mean is None or std_dev is None:
            raise FeatureStoreException(
                "Feature {feature_name:} doesn't have mean and/or standard deviation computed. Thus can't use "
                "standard_scaler method".format(feature_name=feature_name)
            )
        return mean, std_dev

    @staticmethod
    def robust_scaler_stats(
        feature_descriptive_stats: List[FeatureDescriptiveStatistics], feature_name: str
    ):
        percentiles = None
        for stats in feature_descriptive_stats:
            if stats.feature_name == feature_name:
                if stats.feature_type not in ["Integral", "Fractional", "Decimal"]:
                    raise ValueError("Can't compute robust_scaler for this type")
                if stats.percentiles is not None and len(stats.percentiles) > 0:
                    percentiles = stats.percentiles

        if percentiles is None:
            raise FeatureStoreException(
                "Feature {feature_name:} doesn't have mean and/or standard deviation computed. Thus can't use "
                "standard_scaler method".format(feature_name=feature_name)
            )
        return percentiles

    @staticmethod
    def encoder_stats(
        feature_descriptive_stats: List[FeatureDescriptiveStatistics], feature_name: str
    ):
        for stats in feature_descriptive_stats:
            if (
                stats.feature_name == feature_name
                and stats.extended_statistics is not None
                and "unique_values" in stats.extended_statistics
            ):
                unique_data = [
                    value for value in stats.extended_statistics["unique_values"]
                ]
                value_to_index = dict(
                    (value, index) for index, value in enumerate(unique_data)
                )
                return value_to_index



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from hsfs import client
from hsfs.feature_store import FeatureStore


class FeatureStoreApi:
    def __init__(self):
        pass

    def get(self, identifier):
        """Get feature store with specific id or name.

        :param identifier: id or name of the feature store
        :type identifier: int, str
        :return: the featurestore metadata
        :rtype: FeatureStore
        """
        _client = client.get_instance()
        path_params = ["project", _client._project_id, "featurestores", identifier]
        return FeatureStore.from_response_json(
            _client._send_request("GET", path_params)
        )



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import asyncio
import io
import logging
import re

import avro.io
import avro.schema
import numpy as np
import pandas as pd
from hsfs import client, feature_view, training_dataset, util
from hsfs.core import (
    feature_view_api,
    feature_view_engine,
    online_store_rest_client_engine,
    storage_connector_api,
    training_dataset_api,
    transformation_function_engine,
)
from hsfs.serving_key import ServingKey
from sqlalchemy import bindparam, exc, sql, text


_logger = logging.getLogger(__name__)


class VectorServer:
    DEFAULT_ONLINE_STORE_REST_CLIENT = "rest"
    DEFAULT_ONLINE_STORE_SQL_CLIENT = "sql"
    DEFAULT_ONLINE_STORE_CLIENT_KEY = "default_online_store_client"
    ONLINE_REST_CLIENT_CONFIG_OPTIONS_KEY = "config_online_store_rest_client"
    RESET_ONLINE_REST_CLIENT_OPTIONS_KEY = "reset_online_store_rest_client"

    def __init__(
        self,
        feature_store_id,
        features=None,
        training_dataset_version=None,
        serving_keys=None,
        skip_fg_ids=None,
        feature_store_name: str = None,
        feature_view_name: str = None,
        feature_view_version: int = None,
    ):
        if features is None:
            features = []
        self._training_dataset_version = training_dataset_version
        self._features = [] if features is None else features
        self._feature_vector_col_name = (
            [
                feat.name
                for feat in features
                if not (
                    feat.label
                    or feat.inference_helper_column
                    or feat.training_helper_column
                )
            ]
            if features
            else []
        )
        self._skip_fg_ids = skip_fg_ids or set()
        self._prepared_statement_engine = None
        self._prepared_statements = None
        self._helper_column_prepared_statements = None
        self._serving_keys = serving_keys
        self._valid_serving_key = None
        self._pkname_by_serving_index = None
        self._prefix_by_serving_index = None
        self._serving_key_by_serving_index = {}
        self._external = True
        self._feature_store_id = feature_store_id
        self._feature_store_name = feature_store_name
        self._feature_view_name = feature_view_name
        self._feature_view_version = feature_view_version
        self._training_dataset_api = training_dataset_api.TrainingDatasetApi(
            feature_store_id
        )
        self._feature_view_api = feature_view_api.FeatureViewApi(feature_store_id)
        self._storage_connector_api = storage_connector_api.StorageConnectorApi()
        self._transformation_function_engine = (
            transformation_function_engine.TransformationFunctionEngine(
                feature_store_id
            )
        )
        self._feature_view_engine = feature_view_engine.FeatureViewEngine(
            feature_store_id
        )
        self._transformation_functions = None
        self._required_serving_keys = None
        self._async_pool = None

        self._online_store_rest_client_engine = None
        self._init_online_store_rest_client = None
        self._init_online_store_sql_client = None
        self._default_online_store_client = None

    def init_serving(
        self,
        entity: "feature_view.FeatureView",
        batch: bool,
        external: bool = None,
        inference_helper_columns: bool = False,
        init_online_store_sql_client: bool = True,
        init_online_store_rest_client: bool = False,
        options=None,
    ):
        self._set_default_online_store_client(
            init_online_store_rest_client=init_online_store_rest_client,
            init_online_store_sql_client=init_online_store_sql_client,
            options=options,
        )

        if external is None:
            external = isinstance(client.get_instance(), client.external.Client)
        # `init_prepared_statement` should be the last because other initialisations
        # has to be done successfully before it is able to fetch feature vectors.
        self.init_transformation(entity)
        self._complex_features = self.get_complex_feature_schemas()

        if self._init_online_store_rest_client is True:
            _logger.info("Initialising Vector Server Online Store REST client")
            self._online_store_rest_client_engine = online_store_rest_client_engine.OnlineStoreRestClientEngine(
                feature_store_name=self._feature_store_name,
                feature_view_name=entity.name,
                feature_view_version=entity.version,
                features=entity.features,
                skip_fg_ids=self._skip_fg_ids,
                # Code duplication added to avoid unnecessary transforming and iterating over feature vectors
                # multiple times. This is a temporary solution until the code is refactored with new sql client
                complex_features=self._complex_features,
                transformation_functions=self._transformation_functions,
                # Temporary fix to allow for missing primary keys in the entry
                serving_keys=self._serving_keys,
            )
            reset_online_rest_client = False
            online_store_rest_client_config = None
            if isinstance(options, dict):
                reset_online_rest_client = options.get(
                    self.RESET_ONLINE_REST_CLIENT_OPTIONS_KEY, reset_online_rest_client
                )
                online_store_rest_client_config = options.get(
                    self.ONLINE_REST_CLIENT_CONFIG_OPTIONS_KEY,
                    online_store_rest_client_config,
                )

            client.online_store_rest_client.init_or_reset_online_store_rest_client(
                optional_config=online_store_rest_client_config,
                reset_client=reset_online_rest_client,
            )

        if self._init_online_store_sql_client is True:
            _logger.info("Initialising Vector Server Online SQL client")
            self.init_prepared_statement(
                entity,
                batch,
                external,
                inference_helper_columns,
                options=options,
            )

    def init_batch_scoring(self, entity):
        self.init_transformation(entity)

    def init_transformation(self, entity):
        # attach transformation functions
        self._transformation_functions = self._get_transformation_fns(entity)

    def init_prepared_statement(
        self,
        entity,
        batch,
        external,
        inference_helper_columns,
        options=None,
    ):
        if isinstance(entity, feature_view.FeatureView):
            if inference_helper_columns:
                helper_column_prepared_statements = (
                    self._feature_view_api.get_serving_prepared_statement(
                        entity.name, entity.version, batch, inference_helper_columns
                    )
                )
            prepared_statements = self._feature_view_api.get_serving_prepared_statement(
                entity.name, entity.version, batch, inference_helper_columns=False
            )
        elif isinstance(entity, training_dataset.TrainingDataset):
            prepared_statements = (
                self._training_dataset_api.get_serving_prepared_statement(entity, batch)
            )
        else:
            raise ValueError(
                "Object type needs to be `feature_view.FeatureView` or `training_dataset.TrainingDataset`."
            )
        # reset values to default, as user may be re-initialising with different parameters
        self._prepared_statement_engine = None
        self._prepared_statements = None
        self._helper_column_prepared_statements = None
        self._external = external
        # create aiomysql engine
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._set_aiomysql_connection(len(prepared_statements), options=options)
        )

        (
            self._prepared_statements,
            feature_name_order_by_psp,
            prefix_by_serving_index,
        ) = self._parametrize_prepared_statements(prepared_statements, batch)

        if inference_helper_columns:
            (
                self._helper_column_prepared_statements,
                _,
                _,
            ) = self._parametrize_prepared_statements(
                helper_column_prepared_statements, batch
            )

        self._prefix_by_serving_index = prefix_by_serving_index
        for sk in self._serving_keys:
            self._serving_key_by_serving_index[
                sk.join_index
            ] = self._serving_key_by_serving_index.get(sk.join_index, []) + [sk]
        # sort the serving by PreparedStatementParameter.index
        for join_index in self._serving_key_by_serving_index:
            # feature_name_order_by_psp do not include the join index when the joint feature only contains label only
            # But _serving_key_by_serving_index include the index when the join_index is 0 (left side)
            if join_index in feature_name_order_by_psp:
                self._serving_key_by_serving_index[join_index] = sorted(
                    self._serving_key_by_serving_index[join_index],
                    key=lambda _sk: feature_name_order_by_psp[join_index].get(
                        _sk.feature_name, 0
                    ),
                )

        self._valid_serving_key = set(
            [sk.feature_name for sk in self._serving_keys]
            + [sk.required_serving_key for sk in self._serving_keys]
        )

    def _parametrize_prepared_statements(self, prepared_statements, batch):
        prepared_statements_dict = {}
        serving_keys = set()
        feature_name_order_by_psp = dict()
        prefix_by_serving_index = {}
        for prepared_statement in prepared_statements:
            if prepared_statement.feature_group_id in self._skip_fg_ids:
                continue
            query_online = str(prepared_statement.query_online).replace("\n", " ")
            prefix_by_serving_index[
                prepared_statement.prepared_statement_index
            ] = prepared_statement.prefix

            # In java prepared statement `?` is used for parametrization.
            # In sqlalchemy `:feature_name` is used instead of `?`
            pk_names = [
                psp.name
                for psp in sorted(
                    prepared_statement.prepared_statement_parameters,
                    key=lambda psp: psp.index,
                )
            ]
            feature_name_order_by_psp[
                prepared_statement.prepared_statement_index
            ] = dict(
                [
                    (psp.name, psp.index)
                    for psp in prepared_statement.prepared_statement_parameters
                ]
            )
            # construct serving key if it is not provided.
            if self._serving_keys is None:
                for pk_name in pk_names:
                    # should use `ignore_prefix=True` because before hsfs 3.3,
                    # only primary key of a feature group are matched in
                    # user provided entry.
                    serving_key = ServingKey(
                        feature_name=pk_name,
                        join_index=prepared_statement.prepared_statement_index,
                        prefix=prepared_statement.prefix,
                        ignore_prefix=True,
                    )
                    serving_keys.add(serving_key)

            if not batch:
                for pk_name in pk_names:
                    query_online = self._parametrize_query(pk_name, query_online)
                query_online = sql.text(query_online)
            else:
                query_online = self._parametrize_query("batch_ids", query_online)
                query_online = sql.text(query_online)
                query_online = query_online.bindparams(
                    batch_ids=bindparam("batch_ids", expanding=True)
                )

            prepared_statements_dict[
                prepared_statement.prepared_statement_index
            ] = query_online

        # assign serving key if it is not provided.
        if self._serving_keys is None:
            self._serving_keys = serving_keys

        return (
            prepared_statements_dict,
            feature_name_order_by_psp,
            prefix_by_serving_index,
        )

    def _validate_serving_key(self, entry):
        for key in entry:
            if key not in self._valid_serving_key:
                raise ValueError(
                    f"'{key}' is not a correct serving key. Expect one of the"
                    f" followings: [{', '.join(self._valid_serving_key)}]"
                )

    def get_feature_vector(
        self,
        entry,
        return_type=None,
        passed_features=None,
        allow_missing=False,
        force_rest_client: bool = False,
        force_sql_client: bool = False,
    ):
        """Assembles serving vector from online feature store."""
        if passed_features is None:
            passed_features = []

        online_client_str = self.which_client_and_ensure_initialised(
            force_rest_client=force_rest_client, force_sql_client=force_sql_client
        )

        if online_client_str == self.DEFAULT_ONLINE_STORE_REST_CLIENT:
            _logger.info("get_feature_vector Online REST client")
            vector = self._online_store_rest_client_engine.get_single_feature_vector(
                entry=entry,
                passed_features=passed_features,
                return_type=self._online_store_rest_client_engine.RETURN_TYPE_FEATURE_VALUE_LIST,
            )

        else:  # aiomysql branch
            # get result row
            _logger.info("get_feature_vector Online SQL client")
            serving_vector = self._vector_result(entry, self._prepared_statements)
            # Add the passed features
            serving_vector.update(passed_features)

            # apply transformation functions
            result_dict = self._apply_transformation(serving_vector)

            vector = self._generate_vector(result_dict, allow_missing)

        if return_type.lower() == "list":
            return vector
        elif return_type.lower() == "numpy":
            return np.array(vector)
        elif return_type.lower() == "pandas":
            pandas_df = pd.DataFrame(vector).transpose()
            pandas_df.columns = self._feature_vector_col_name
            return pandas_df
        else:
            raise ValueError(
                "Unknown return type. Supported return types are 'list','pandas' and 'numpy'"
            )

    def get_feature_vectors(
        self,
        entries,
        return_type=None,
        passed_features=None,
        allow_missing=False,
        force_rest_client: bool = False,
        force_sql_client: bool = False,
    ):
        if passed_features is None:
            passed_features = []
        """Assembles serving vector from online feature store."""
        online_client_str = self.which_client_and_ensure_initialised(
            force_rest_client=force_rest_client, force_sql_client=force_sql_client
        )

        if online_client_str == self.DEFAULT_ONLINE_STORE_REST_CLIENT:
            _logger.info("get_feature_vectors through REST client")
            vectors = self._online_store_rest_client_engine.get_batch_feature_vectors(
                entries=entries,
                passed_features=passed_features,
                return_type=self._online_store_rest_client_engine.RETURN_TYPE_FEATURE_VALUE_LIST,
            )
        else:
            _logger.info("get_feature_vectors through SQL client")
            batch_results, _ = self._batch_vector_results(
                entries, self._prepared_statements
            )
            # apply passed features to each batch result
            for vector_index, pf in enumerate(passed_features):
                batch_results[vector_index].update(pf)

            # apply transformation functions
            batch_transformed = list(
                map(
                    lambda results_dict: self._apply_transformation(results_dict),
                    batch_results,
                )
            )

            # get vectors
            vectors = []
            for result in batch_transformed:
                # for backward compatibility, before 3.4, if result is empty,
                # instead of throwing error, it skips the result
                if len(result) != 0 or allow_missing:
                    vectors.append(self._generate_vector(result, fill_na=allow_missing))

        if return_type.lower() == "list":
            return vectors
        elif return_type.lower() == "numpy":
            return np.array(vectors)
        elif return_type.lower() == "pandas":
            pandas_df = pd.DataFrame(vectors)
            pandas_df.columns = self._feature_vector_col_name
            return pandas_df
        else:
            raise ValueError(
                "Unknown return type. Supported return types are 'list', 'pandas' and 'numpy'"
            )

    def which_client_and_ensure_initialised(
        self, force_rest_client: bool, force_sql_client: bool
    ) -> str:
        """Check if the requested client is initialised as well as deciding which client to use based on default.

        # Arguments:
            force_rest_client: bool. user specified override to use rest_client.
            force_sql_client: bool. user specified override to use sql_client.

        # Returns:
            An enum specifying the client to be used.
        """
        if force_rest_client and force_sql_client:
            raise ValueError(
                "force_rest_client and force_sql_client cannot be used at the same time."
            )

        # No override, use default client
        if not force_rest_client and not force_sql_client:
            return self.default_online_store_client

        if (
            self._init_online_store_rest_client is False
            and self._init_online_store_sql_client is False
        ):
            raise ValueError(
                "No client is initialised. Call `init_serving` with init_online_store_sql_client or init_online_store_rest_client set to True before using it."
            )
        if force_sql_client and (self._init_online_store_sql_client is False):
            raise ValueError(
                "SQL Client is not initialised. Call `init_serving` with init_online_store_sql_client set to True before using it."
            )
        elif force_sql_client:
            return self.DEFAULT_ONLINE_STORE_SQL_CLIENT

        if force_rest_client and (self._init_online_store_rest_client is False):
            raise ValueError(
                "RonDB Rest Client is not initialised. Call `init_serving` with init_online_store_rest_client set to True before using it."
            )
        elif force_rest_client:
            return self.DEFAULT_ONLINE_STORE_REST_CLIENT

    def _set_default_online_store_client(
        self,
        init_online_store_rest_client: bool,
        init_online_store_sql_client: bool,
        options: dict,
    ):
        if (
            init_online_store_rest_client is False
            and init_online_store_sql_client is False
        ):
            raise ValueError(
                "At least one of the clients should be initialised. Set init_online_store_sql_client or init_online_store_rest_client to True."
            )
        self._init_online_store_rest_client = init_online_store_rest_client
        self._init_online_store_sql_client = init_online_store_sql_client

        if (
            init_online_store_rest_client is True
            and init_online_store_sql_client is True
        ):
            # Defaults to SQL as client for legacy reasons mainly.
            self.default_online_store_client = (
                options.get(
                    "default_online_store_client", self.DEFAULT_ONLINE_STORE_SQL_CLIENT
                )
                if isinstance(options, dict)
                else self.DEFAULT_ONLINE_STORE_SQL_CLIENT
            )
        elif init_online_store_rest_client is True:
            self.default_online_store_client = self.DEFAULT_ONLINE_STORE_REST_CLIENT
        else:
            self.default_online_store_client = self.DEFAULT_ONLINE_STORE_SQL_CLIENT

    def get_inference_helper(self, entry, return_type):
        """Assembles serving vector from online feature store."""

        serving_vector = self._vector_result(
            entry, self._helper_column_prepared_statements
        )

        if return_type.lower() == "pandas":
            return pd.DataFrame([serving_vector])
        elif return_type.lower() == "dict":
            return serving_vector
        else:
            raise ValueError(
                "Unknown return type. Supported return types are 'pandas' and 'dict'"
            )

    def get_inference_helpers(
        self,
        feature_view_object,
        entries,
        return_type,
    ):
        """Assembles serving vector from online feature store."""
        batch_results, serving_keys = self._batch_vector_results(
            entries, self._helper_column_prepared_statements
        )

        # drop serving and primary key names from the result dict
        drop_list = serving_keys + list(feature_view_object.primary_keys)
        _ = list(
            map(
                lambda results_dict: [
                    results_dict.pop(x, None)
                    for x in drop_list
                    if x not in feature_view_object.inference_helper_columns
                ],
                batch_results,
            )
        )

        if return_type.lower() == "dict":
            return batch_results
        elif return_type.lower() == "pandas":
            return pd.DataFrame(batch_results)
        else:
            raise ValueError(
                "Unknown return type. Supported return types are 'dict' and 'pandas'"
            )

    def _vector_result(self, entry, prepared_statement_objects):
        """Retrieve single vector with parallel queries using aiomysql engine."""

        if all([isinstance(val, list) for val in entry.values()]):
            raise ValueError(
                "Entry is expected to be single value per primary key. "
                "If you have already initialised prepared statements for single vector and now want to retrieve "
                "batch vector please reinitialise prepared statements with  "
                "`training_dataset.init_prepared_statement()` "
                "or `feature_view.init_serving()`"
            )
        self._validate_serving_key(entry)
        # Initialize the set of values
        serving_vector = {}
        bind_entries = {}
        prepared_statement_execution = {}
        for prepared_statement_index in prepared_statement_objects:
            pk_entry = {}
            next_statement = False
            for sk in self._serving_key_by_serving_index[prepared_statement_index]:
                if sk.required_serving_key not in entry.keys():
                    # Check if there is any entry matched with feature name.
                    if sk.feature_name in entry.keys():
                        pk_entry[sk.feature_name] = entry[sk.feature_name]
                    else:
                        # User did not provide the necessary serving keys, we expect they have
                        # provided the necessary features as passed_features.
                        # We are going to check later if this is true
                        next_statement = True
                        break
                else:
                    pk_entry[sk.feature_name] = entry[sk.required_serving_key]
            if next_statement:
                continue
            bind_entries[prepared_statement_index] = pk_entry
            prepared_statement_execution[
                prepared_statement_index
            ] = prepared_statement_objects[prepared_statement_index]

        # run all the prepared statements in parallel using aiomysql engine
        loop = asyncio.get_event_loop()
        results_dict = loop.run_until_complete(
            self._execute_prep_statements(prepared_statement_execution, bind_entries)
        )

        for key in results_dict:
            for row in results_dict[key]:
                result_dict = self.deserialize_complex_features(dict(row))
                serving_vector.update(result_dict)

        return serving_vector

    def _batch_vector_results(self, entries, prepared_statement_objects):
        """Execute prepared statements in parallel using aiomysql engine."""

        # create dict object that will have of order of the vector as key and values as
        # vector itself to stitch them correctly if there are multiple feature groups involved. At this point we
        # expect that backend will return correctly ordered vectors.
        batch_results = [{} for _ in range(len(entries))]
        entry_values = {}
        serving_keys_all_fg = []
        prepared_stmts_to_execute = {}
        # construct the list of entry values for binding to query
        for prepared_statement_index in prepared_statement_objects:
            # prepared_statement_index include fg with label only
            # But _serving_key_by_serving_index include the index when the join_index is 0 (left side)
            if prepared_statement_index not in self._serving_key_by_serving_index:
                continue

            prepared_stmts_to_execute[
                prepared_statement_index
            ] = prepared_statement_objects[prepared_statement_index]
            entry_values_tuples = list(
                map(
                    lambda e: tuple(
                        [
                            (
                                e.get(sk.required_serving_key)
                                # Check if there is any entry matched with feature name,
                                # if the required serving key is not provided.
                                or e.get(sk.feature_name)
                            )
                            for sk in self._serving_key_by_serving_index[
                                prepared_statement_index
                            ]
                        ]
                    ),
                    entries,
                )
            )
            entry_values[prepared_statement_index] = {"batch_ids": entry_values_tuples}

        # run all the prepared statements in parallel using aiomysql engine
        loop = asyncio.get_event_loop()
        parallel_results = loop.run_until_complete(
            self._execute_prep_statements(prepared_stmts_to_execute, entry_values)
        )

        # construct the results
        for prepared_statement_index in prepared_stmts_to_execute:
            statement_results = {}
            serving_keys = self._serving_key_by_serving_index[prepared_statement_index]
            serving_keys_all_fg += serving_keys
            # Use prefix from prepare statement because prefix from serving key is collision adjusted.
            prefix_features = [
                (self._prefix_by_serving_index[prepared_statement_index] or "")
                + sk.feature_name
                for sk in self._serving_key_by_serving_index[prepared_statement_index]
            ]
            # iterate over results by index of the prepared statement
            for row in parallel_results[prepared_statement_index]:
                row_dict = dict(row)
                # can primary key be complex feature?
                result_dict = self.deserialize_complex_features(row_dict)
                # note: should used serialized value
                # as it is from users' input
                statement_results[
                    self._get_result_key(prefix_features, row_dict)
                ] = result_dict

            # add partial results to the global results
            for i, entry in enumerate(entries):
                batch_results[i].update(
                    statement_results.get(
                        self._get_result_key_serving_key(serving_keys, entry), {}
                    )
                )
        return batch_results, serving_keys_all_fg

    def get_complex_feature_schemas(self):
        return {
            f.name: avro.io.DatumReader(
                avro.schema.parse(
                    f._feature_group._get_feature_avro_schema(
                        f.feature_group_feature_name
                    )
                )
            )
            for f in self._features
            if f.is_complex()
        }

    def deserialize_complex_features(self, row_dict):
        for feature_name, schema in self._complex_features.items():
            if feature_name in row_dict:
                bytes_reader = io.BytesIO(row_dict[feature_name])
                decoder = avro.io.BinaryDecoder(bytes_reader)
                row_dict[feature_name] = schema.read(decoder)
        return row_dict

    def refresh_mysql_connection(self):
        try:
            with self._prepared_statement_engine.connect():
                pass
        except exc.OperationalError:
            self._set_mysql_connection()

    def _make_preview_statement(self, statement, n):
        return text(statement.text[: statement.text.find(" WHERE ")] + f" LIMIT {n}")

    def _set_mysql_connection(self, options=None):
        online_conn = self._storage_connector_api.get_online_connector(
            self._feature_store_id
        )
        self._prepared_statement_engine = util.create_mysql_engine(
            online_conn, self._external, options=options
        )

    def _generate_vector(self, result_dict, fill_na=False):
        # feature values
        vector = []
        for feature_name in self._feature_vector_col_name:
            if feature_name not in result_dict:
                if fill_na:
                    vector.append(None)
                else:
                    raise client.exceptions.FeatureStoreException(
                        f"Feature '{feature_name}' is missing from vector."
                        "Possible reasons: "
                        "1. There is no match in the given entry."
                        " Please check if the entry exists in the online feature store"
                        " or provide the feature as passed_feature. "
                        f"2. Required entries [{', '.join(self.required_serving_keys)}] or "
                        f"[{', '.join(set(sk.feature_name for sk in self._serving_keys))}] are not provided."
                    )
            else:
                vector.append(result_dict[feature_name])
        return vector

    def _apply_transformation(self, row_dict):
        for feature_name in self._transformation_functions:
            if feature_name in row_dict:
                transformation_fn = self._transformation_functions[
                    feature_name
                ].transformation_fn
                row_dict[feature_name] = transformation_fn(row_dict[feature_name])
        return row_dict

    @staticmethod
    def _get_result_key(primary_keys, result_dict):
        result_key = []
        for pk in primary_keys:
            result_key.append(result_dict.get(pk))
        return tuple(result_key)

    @staticmethod
    def _get_result_key_serving_key(serving_keys, result_dict):
        result_key = []
        for sk in serving_keys:
            result_key.append(
                result_dict.get(sk.required_serving_key)
                or result_dict.get(sk.feature_name)
            )
        return tuple(result_key)

    @staticmethod
    def _parametrize_query(name, query_online):
        # Now we have ordered pk_names, iterate over it and replace `?` with `:feature_name` one by one.
        # Regex `"^(.*?)\?"` will identify 1st occurrence of `?` in the sql string and replace it with provided
        # feature name, e.g. `:feature_name`:. As we iteratively update `query_online` we are always aiming to
        # replace 1st occurrence of `?`. This approach can only work if primary key names are sorted properly.
        # Regex `"^(.*?)\?"`:
        # `^` - asserts position at start of a line
        # `.*?` - matches any character (except for line terminators). `*?` Quantifier —
        # Matches between zero and unlimited times, expanding until needed, i.e 1st occurrence of `\?`
        # character.
        return re.sub(
            r"^(.*?)\?",
            r"\1:" + name,
            query_online,
        )

    def _get_transformation_fns(self, entity):
        # get attached transformation functions
        transformation_functions = (
            self._transformation_function_engine.get_td_transformation_fn(entity)
            if isinstance(entity, training_dataset.TrainingDataset)
            else (
                self._feature_view_engine.get_attached_transformation_fn(
                    entity.name, entity.version
                )
            )
        )
        is_stat_required = (
            len(
                set(self._transformation_function_engine.BUILTIN_FN_NAMES).intersection(
                    set([tf.name for tf in transformation_functions.values()])
                )
            )
            > 0
        )
        is_feat_view = isinstance(entity, feature_view.FeatureView)
        if not is_stat_required:
            td_tffn_stats = None
        else:
            # if there are any built-in transformation functions get related statistics and
            # populate with relevant arguments
            # there should be only one statistics object with before_transformation=true
            if is_feat_view and self._training_dataset_version is None:
                raise ValueError(
                    "Training data version is required for transformation. Call `feature_view.init_serving(version)` "
                    "or `feature_view.init_batch_scoring(version)` to pass the training dataset version."
                    "Training data can be created by `feature_view.create_training_data` or `feature_view.get_training_data`."
                )
            td_tffn_stats = self._feature_view_engine._statistics_engine.get(
                entity,
                before_transformation=True,
                training_dataset_version=self._training_dataset_version,
            )

        if is_stat_required and td_tffn_stats is None:
            raise ValueError(
                "No statistics available for initializing transformation functions."
            )

        transformation_fns = (
            self._transformation_function_engine.populate_builtin_attached_fns(
                transformation_functions,
                td_tffn_stats.feature_descriptive_statistics
                if td_tffn_stats is not None
                else None,
            )
        )
        return transformation_fns

    def filter_entry_by_join_index(self, entry, join_index):
        fg_entry = {}
        complete = True
        for sk in self._serving_key_by_serving_index[join_index]:
            fg_entry[sk.feature_name] = entry.get(sk.required_serving_key) or entry.get(
                sk.feature_name
            )  # fallback to use raw feature name
            if fg_entry[sk.feature_name] is None:
                complete = False
                break
        return complete, fg_entry

    async def _set_aiomysql_connection(self, default_min_size: int, options=None):
        online_connector = self._storage_connector_api.get_online_connector(
            self._feature_store_id
        )
        self._async_pool = await util.create_async_engine(
            online_connector, self._external, default_min_size, options=options
        )

    async def _query_async_sql(self, stmt, bind_params):
        """Query prepared statement together with bind params using aiomysql connection pool"""
        # Get a connection from the pool
        async with self._async_pool.acquire() as conn:
            # Execute the prepared statement
            cursor = await conn.execute(stmt, bind_params)
            # Fetch the result
            resultset = await cursor.fetchall()
            await cursor.close()

        return resultset

    async def _execute_prep_statements(self, prepared_statements: dict, entries: dict):
        """Iterate over prepared statements to create async tasks
        and gather all tasks results for a given list of entries."""

        # validate if prepared_statements and entries have the same keys
        if prepared_statements.keys() != entries.keys():
            # iterate over prepared_statements and entries to find the missing key
            # remove missing keys from prepared_statements
            for key in list(prepared_statements.keys()):
                if key not in entries:
                    prepared_statements.pop(key)

        tasks = [
            asyncio.ensure_future(
                self._query_async_sql(prepared_statements[key], entries[key])
            )
            for key in prepared_statements
        ]
        # Run the queries in parallel using asyncio.gather
        results = await asyncio.gather(*tasks)
        results_dict = {}
        # Create a dict of results with the prepared statement index as key
        for i, key in enumerate(prepared_statements):
            results_dict[key] = results[i]
        # TODO: close connection? closing connection pool here will make un-acquirable for future connections
        # unless init_serving is explicitly called before get_feature_vector(s).
        return results_dict

    @property
    def prepared_statement_engine(self):
        """JDBC connection engine to retrieve connections to online features store from."""
        return self._prepared_statement_engine

    @prepared_statement_engine.setter
    def prepared_statement_engine(self, prepared_statement_engine):
        self._prepared_statement_engine = prepared_statement_engine

    @property
    def prepared_statements(self):
        """The dict object of prepared_statements as values and kes as indices of positions in the query for
        selecting features from feature groups of the training dataset.
        """
        return self._prepared_statements

    @prepared_statements.setter
    def prepared_statements(self, prepared_statements):
        self._prepared_statements = prepared_statements

    @property
    def required_serving_keys(self):
        """Set of primary key names that is used as keys in input dict object for `get_feature_vector` method."""
        if self._required_serving_keys is not None:
            return self._required_serving_keys
        if self._serving_keys is not None:
            self._required_serving_keys = set(
                [key.required_serving_key for key in self._serving_keys]
            )
        else:
            self._required_serving_keys = set()
        return self._required_serving_keys

    @property
    def serving_keys(self):
        return self._serving_keys

    @serving_keys.setter
    def serving_keys(self, serving_vector_keys):
        self._serving_keys = serving_vector_keys

    @property
    def training_dataset_version(self):
        return self._training_dataset_version

    @training_dataset_version.setter
    def training_dataset_version(self, training_dataset_version):
        self._training_dataset_version = training_dataset_version

    @property
    def default_online_store_client(self) -> str:
        return self._default_online_store_client

    @default_online_store_client.setter
    def default_online_store_client(self, default_online_store_client: str):
        if default_online_store_client not in [
            self.DEFAULT_ONLINE_STORE_REST_CLIENT,
            self.DEFAULT_ONLINE_STORE_SQL_CLIENT,
        ]:
            raise ValueError(
                f"Default Online Feature Store Client should be one of {self.DEFAULT_ONLINE_STORE_REST_CLIENT} or {self.DEFAULT_ONLINE_STORE_SQL_CLIENT}."
            )

        if (
            default_online_store_client == self.DEFAULT_ONLINE_STORE_REST_CLIENT
            and self._init_online_store_rest_client is False
        ):
            raise ValueError(
                f"Default Online Store cCient is set to {self.DEFAULT_ONLINE_STORE_REST_CLIENT} but Online Store REST client"
                + " is not initialised. Call `init_serving` with init_client set to True before using it."
            )
        elif (
            default_online_store_client == self.DEFAULT_ONLINE_STORE_SQL_CLIENT
            and self._init_online_store_sql_client is False
        ):
            raise ValueError(
                f"Default online client is set to {self.DEFAULT_ONLINE_STORE_SQL_CLIENT} but Online Store SQL client"
                + " is not initialised. Call `init_serving` with init_online_store_sql_client set to True before using it."
            )

        _logger.info(
            f"Default Online Store Client is set to {default_online_store_client}."
        )
        self._default_online_store_client = default_online_store_client



#
#   Copyright 2023 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import datetime
import json
import base64
import warnings
from functools import wraps

import pyarrow
import pyarrow.flight
from pyarrow.flight import FlightServerError
from hsfs import client
from hsfs import feature_group
from hsfs.client.exceptions import FeatureStoreException
from hsfs.core.variable_api import VariableApi
from hsfs import util
from hsfs.storage_connector import StorageConnector
from retrying import retry

_arrow_flight_instance = None


def get_instance():
    global _arrow_flight_instance
    if not _arrow_flight_instance:
        _arrow_flight_instance = ArrowFlightClient()
    return _arrow_flight_instance


def close():
    global _arrow_flight_instance
    _arrow_flight_instance = None


class ArrowFlightClient:
    SUPPORTED_FORMATS = ["parquet"]
    SUPPORTED_EXTERNAL_CONNECTORS = [
        StorageConnector.SNOWFLAKE,
        StorageConnector.BIGQUERY,
    ]
    FILTER_NUMERIC_TYPES = ["bigint", "tinyint", "smallint", "int", "float", "double"]
    READ_ERROR = 'Could not read data using ArrowFlight. If the issue persists, use read_options={"use_hive": True} instead.'
    WRITE_ERROR = 'Could not write data using ArrowFlight. If the issue persists, use write_options={"use_spark": True} instead.'
    DEFAULT_TIMEOUT = 900

    def __init__(self):
        try:
            self._variable_api = VariableApi()
            self._is_enabled = self._variable_api.get_flyingduck_enabled()
        except Exception:
            # if feature flag cannot be retrieved, assume it is disabled
            self._is_enabled = False
            return

        if self._is_enabled:
            try:
                self._initialize_connection()
            except Exception as e:
                self._disable(str(e))

    def _disable(self, message):
        self._is_enabled = False
        warnings.warn(
            f"Could not establish connection to ArrowFlight Server. ({message}) "
            f"Will fall back to hive/spark for this session. "
            f"If the error persists, you can disable using ArrowFlight "
            f"by changing the cluster configuration (set 'enable_flyingduck'='false')."
        )

    def _initialize_connection(self):
        self._client = client.get_instance()

        if isinstance(self._client, client.external.Client):
            external_domain = self._variable_api.get_loadbalancer_external_domain()
            if external_domain == "":
                raise Exception(
                    "External client could not locate loadbalancer_external_domain "
                    "in cluster configuration or variable is empty."
                )
            host_url = f"grpc+tls://{external_domain}:5005"
        else:
            service_discovery_domain = self._variable_api.get_service_discovery_domain()
            if service_discovery_domain == "":
                raise Exception(
                    "Client could not locate service_discovery_domain "
                    "in cluster configuration or variable is empty."
                )
            host_url = f"grpc+tls://flyingduck.service.{service_discovery_domain}:5005"

        (tls_root_certs, cert_chain, private_key) = self._extract_certs(self._client)
        self._connection = pyarrow.flight.FlightClient(
            location=host_url,
            tls_root_certs=tls_root_certs,
            cert_chain=cert_chain,
            private_key=private_key,
            override_hostname="flyingduck.service.consul",
        )
        self._health_check()
        self._register_certificates()

    def _health_check(self):
        action = pyarrow.flight.Action("healthcheck", b"")
        options = pyarrow.flight.FlightCallOptions(timeout=1)
        list(self._connection.do_action(action, options=options))

    def _should_be_used(self, read_options):
        if read_options and (
            read_options.get("use_hive", False) or read_options.get("use_spark", False)
        ):
            return False

        return self._is_enabled

    def is_query_supported(self, query, read_options):
        return self._should_be_used(read_options) and self._is_query_supported_rec(
            query
        )

    def is_data_format_supported(self, data_format, read_options):
        return (
            self._should_be_used(read_options)
            and data_format in ArrowFlightClient.SUPPORTED_FORMATS
        )

    def _is_query_supported_rec(self, query):
        hudi_no_time_travel = (
            isinstance(query._left_feature_group, feature_group.FeatureGroup)
            and query._left_feature_group.time_travel_format == "HUDI"
            and (
                query._left_feature_group_start_time is None
                or query._left_feature_group_start_time == 0
            )
            and query._left_feature_group_end_time is None
        )
        supported_connector = (
            isinstance(query._left_feature_group, feature_group.ExternalFeatureGroup)
            and query._left_feature_group.storage_connector.type
            in ArrowFlightClient.SUPPORTED_EXTERNAL_CONNECTORS
        )
        supported = hudi_no_time_travel or supported_connector
        for j in query._joins:
            supported &= self._is_query_supported_rec(j._query)
        return supported

    def _extract_certs(self, client):
        with open(client._get_ca_chain_path(), "rb") as f:
            tls_root_certs = f.read()
        with open(client._get_client_cert_path(), "r") as f:
            cert_chain = f.read()
        with open(client._get_client_key_path(), "r") as f:
            private_key = f.read()
        return tls_root_certs, cert_chain, private_key

    def _encode_certs(self, path):
        with open(path, "rb") as f:
            content = f.read()
            return base64.b64encode(content).decode("utf-8")

    def _register_certificates(self):
        kstore = self._encode_certs(self._client._get_jks_key_store_path())
        tstore = self._encode_certs(self._client._get_jks_trust_store_path())
        cert_key = self._client._cert_key
        certificates_json = json.dumps(
            {"kstore": kstore, "tstore": tstore, "cert_key": cert_key}
        ).encode("ascii")
        certificates_json_buf = pyarrow.py_buffer(certificates_json)
        action = pyarrow.flight.Action(
            "register-client-certificates", certificates_json_buf
        )
        self._connection.do_action(action)

    def _handle_afs_exception(user_message="None"):
        def decorator(func):
            @wraps(func)
            def afs_error_handler_wrapper(instance, *args, **kw):
                try:
                    return func(instance, *args, **kw)
                except Exception as e:
                    message = str(e)
                    if (
                        isinstance(e, FlightServerError)
                        and "Please register client certificates first." in message
                    ):
                        instance._register_certificates()
                        return func(instance, *args, **kw)
                    else:
                        raise FeatureStoreException(user_message) from e

            return afs_error_handler_wrapper

        return decorator

    def _should_retry(exception):
        return isinstance(exception, pyarrow._flight.FlightUnavailableError)

    @retry(
        wait_exponential_multiplier=1000,
        stop_max_attempt_number=5,
        retry_on_exception=_should_retry,
    )
    def get_flight_info(self, descriptor):
        return self._connection.get_flight_info(descriptor)

    def _get_dataset(self, descriptor, timeout=DEFAULT_TIMEOUT):
        info = self.get_flight_info(descriptor)
        options = pyarrow.flight.FlightCallOptions(timeout=timeout)
        reader = self._connection.do_get(self._info_to_ticket(info), options)
        return reader.read_pandas()

    @_handle_afs_exception(user_message=READ_ERROR)
    def read_query(self, query_object, arrow_flight_config):
        query_encoded = json.dumps(query_object).encode("ascii")
        descriptor = pyarrow.flight.FlightDescriptor.for_command(query_encoded)
        return self._get_dataset(
            descriptor,
            arrow_flight_config.get("timeout")
            if arrow_flight_config
            else self.DEFAULT_TIMEOUT,
        )

    @_handle_afs_exception(user_message=READ_ERROR)
    def read_path(self, path, arrow_flight_config):
        descriptor = pyarrow.flight.FlightDescriptor.for_path(path)
        return self._get_dataset(descriptor, arrow_flight_config)

    @_handle_afs_exception(user_message=WRITE_ERROR)
    def create_training_dataset(
        self, feature_view_obj, training_dataset_obj, query_obj, arrow_flight_config
    ):
        training_dataset = {}
        training_dataset["fs_name"] = util.strip_feature_store_suffix(
            training_dataset_obj.feature_store_name
        )
        training_dataset["fv_name"] = feature_view_obj.name
        training_dataset["fv_version"] = feature_view_obj.version
        training_dataset["tds_version"] = training_dataset_obj.version
        training_dataset["query"] = query_obj

        try:
            training_dataset_encoded = json.dumps(training_dataset).encode("ascii")
            training_dataset_buf = pyarrow.py_buffer(training_dataset_encoded)
            action = pyarrow.flight.Action(
                "create-training-dataset", training_dataset_buf
            )
            timeout = (
                arrow_flight_config.get("timeout", self.DEFAULT_TIMEOUT)
                if arrow_flight_config
                else self.DEFAULT_TIMEOUT
            )
            options = pyarrow.flight.FlightCallOptions(timeout=timeout)
            for result in self._connection.do_action(action, options):
                return result.body.to_pybytes()
        except pyarrow.lib.ArrowIOError as e:
            print("Error calling action:", e)

    def is_flyingduck_query_object(self, query_obj):
        return isinstance(query_obj, dict) and "query_string" in query_obj

    def create_query_object(self, query, query_str, on_demand_fg_aliases=[]):
        features = {}
        connectors = {}
        for fg in query.featuregroups:
            fg_name = self._serialize_featuregroup_name(fg)
            fg_connector = self._serialize_featuregroup_connector(
                fg, query, on_demand_fg_aliases
            )
            features[fg_name] = [feat.name for feat in fg.features]
            connectors[fg_name] = fg_connector
        filters = self._serialize_filter_expression(query.filters, query)

        query = {
            "query_string": self._translate_to_duckdb(query, query_str),
            "features": features,
            "filters": filters,
            "connectors": connectors,
        }
        return query

    def _serialize_featuregroup_connector(self, fg, query, on_demand_fg_aliases):
        connector = {}
        if isinstance(fg, feature_group.ExternalFeatureGroup):
            connector["type"] = fg.storage_connector.type
            connector["options"] = fg.storage_connector.connector_options()
            connector["query"] = fg.query[:-1] if fg.query.endswith(";") else fg.query
            for on_demand_fg_alias in on_demand_fg_aliases:
                if on_demand_fg_alias.on_demand_feature_group.name == fg.name:
                    connector["alias"] = on_demand_fg_alias.alias
                    break
            if query._left_feature_group == fg:
                connector["filters"] = self._serialize_filter_expression(
                    query._filter, query, True
                )
            else:
                for join_obj in query._joins:
                    if join_obj._query._left_feature_group == fg:
                        connector["filters"] = self._serialize_filter_expression(
                            join_obj._query._filter, join_obj._query, True
                        )
        else:
            connector["type"] = "hudi"

        return connector

    def _serialize_featuregroup_name(self, fg):
        return f"{fg._get_project_name()}.{fg.name}_{fg.version}"

    def _serialize_filter_expression(self, filters, query, short_name=False):
        if filters is None:
            return None
        return self._serialize_logic(filters, query, short_name)

    def _serialize_logic(self, logic, query, short_name):
        return {
            "type": "logic",
            "logic_type": logic._type,
            "left_filter": self._serialize_filter_or_logic(
                logic._left_f, logic._left_l, query, short_name
            ),
            "right_filter": self._serialize_filter_or_logic(
                logic._right_f, logic._right_l, query, short_name
            ),
        }

    def _serialize_filter_or_logic(self, filter, logic, query, short_name):
        if filter:
            return self._serialize_filter(filter, query, short_name)
        elif logic:
            return self._serialize_logic(logic, query, short_name)
        else:
            return None

    def _serialize_filter(self, filter, query, short_name):
        if isinstance(filter._value, datetime.datetime):
            filter_value = filter._value.strftime("%Y-%m-%d %H:%M:%S")
        else:
            filter_value = filter._value

        return {
            "type": "filter",
            "condition": filter._condition,
            "value": filter_value,
            "feature": self._serialize_feature_name(filter._feature, query, short_name),
        }

    def _serialize_feature_name(self, feature, query, short_name):
        if short_name:
            return feature.name

        fg = query._get_featuregroup_by_feature(feature)
        fg_name = self._serialize_featuregroup_name(fg)
        return f"{fg_name}.{feature.name}"

    def _translate_to_duckdb(self, query, query_str):
        translated = query_str
        for fg in query.featuregroups:
            translated = translated.replace(
                f"`{fg.feature_store_name}`.`",
                f"`{fg._get_project_name()}.",
            )
        return translated.replace("`", '"')

    def _info_to_ticket(self, info):
        return info.endpoints[0].ticket

    def is_enabled(self):
        return self._is_enabled

    def supports(self, featuregroups):
        if len(featuregroups) > sum(
            1
            for fg in featuregroups
            if isinstance(
                fg,
                (feature_group.FeatureGroup, feature_group.ExternalFeatureGroup),
            )
        ):
            # Contains unsupported feature group types such as a spine group
            return False

        for fg in filter(
            lambda fg: isinstance(fg, feature_group.ExternalFeatureGroup), featuregroups
        ):
            if fg.storage_connector.type not in self.SUPPORTED_EXTERNAL_CONNECTORS:
                return False
        return True



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import humps

import hsfs.feature_group as feature_group
from hsfs.feature import Feature
from hsfs.transformation_function import TransformationFunction


class TrainingDatasetFeature:
    def __init__(
        self,
        name,
        type=None,
        index=None,
        featuregroup=None,
        feature_group_feature_name=None,
        label=False,
        inference_helper_column=False,
        training_helper_column=False,
        transformation_function=None,
        **kwargs,
    ):
        self._name = name.lower()
        self._type = type
        self._index = index
        self._feature_group = (
            feature_group.FeatureGroup.from_response_json(featuregroup)
            if isinstance(featuregroup, dict)
            else featuregroup
        )
        self._feature_group_feature_name = feature_group_feature_name
        self._label = label
        self._inference_helper_column = inference_helper_column
        self._training_helper_column = training_helper_column
        self._transformation_function = (
            TransformationFunction.from_response_json(transformation_function)
            if isinstance(transformation_function, dict)
            else transformation_function
        )

    def to_dict(self):
        return {
            "name": self._name,
            "type": self._type,
            "index": self._index,
            "label": self._label,
            "inferenceHelperColumn": self._inference_helper_column,
            "trainingHelperColumn": self._training_helper_column,
            "transformationFunction": self._transformation_function,
            "featureGroupFeatureName": self._feature_group_feature_name,
            "featuregroup": self._feature_group,
        }

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        return cls(**json_decamelized)

    def is_complex(self):
        """Returns true if the feature has a complex type."""
        return any(map(self._type.upper().startswith, Feature.COMPLEX_TYPES))

    @property
    def name(self):
        """Name of the feature."""
        return self._name

    @property
    def type(self):
        """Data type of the feature in the feature store.

        !!! danger "Not a Python type"
            This type property is not to be confused with Python types.
            The type property represents the actual data type of the feature in
            the feature store.
        """
        return self._type

    @property
    def index(self):
        """Index of the feature in the training dataset, required to restore the correct
        order of features."""
        return self._index

    @property
    def label(self):
        """Indicator if the feature is part of the prediction label."""
        return self._label

    @label.setter
    def label(self, label):
        self._label = label

    @property
    def inference_helper_column(self):
        """Indicator if it is feature."""
        return self._inference_helper_column

    @inference_helper_column.setter
    def inference_helper_column(self, inference_helper_column):
        self._inference_helper_column = inference_helper_column

    @property
    def training_helper_column(self):
        """Indicator if it is feature."""
        return self._training_helper_column

    @training_helper_column.setter
    def training_helper_column(self, training_helper_column):
        self._training_helper_column = training_helper_column

    @property
    def transformation_function(self):
        """Set transformation functions."""
        return self._transformation_function

    @transformation_function.setter
    def transformation_function(self, transformation_function):
        self._transformation_function = transformation_function

    @property
    def feature_group(self):
        return self._feature_group

    @property
    def feature_group_feature_name(self):
        return self._feature_group_feature_name

    def __repr__(self):
        return f"Training Dataset Feature({self._name!r}, {self._type!r}, {self._index!r}, {self._label}, {self._transformation_function}, {self._feature_group_feature_name}, {self._feature_group.id!r})"



#
#   Copyright 2024 Hopsworks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import humps
import json
from datetime import datetime, timezone
from hsfs import util


class JobSchedule:
    def __init__(
        self,
        start_date_time,
        enabled,
        cron_expression,
        next_execution_date_time=None,
        id=None,
        end_date_time=None,
        **kwargs,
    ):
        self._id = id
        self._start_date_time = (
            datetime.fromtimestamp(start_date_time / 1000, tz=timezone.utc)
            if isinstance(start_date_time, int)
            else start_date_time
        )

        self._end_date_time = (
            datetime.fromtimestamp(end_date_time / 1000, tz=timezone.utc)
            if isinstance(end_date_time, int)
            else end_date_time
        )
        self._enabled = enabled
        self._cron_expression = cron_expression

        self._next_execution_date_time = (
            datetime.fromtimestamp(next_execution_date_time / 1000, tz=timezone.utc)
            if isinstance(next_execution_date_time, int)
            else next_execution_date_time
        )

    @classmethod
    def from_response_json(cls, json_dict):
        json_decamelized = humps.decamelize(json_dict)
        return cls(**json_decamelized)

    def to_dict(self):
        return {
            "id": self._id,
            "startDateTime": int(self._start_date_time.timestamp() * 1000.0)
            if self._start_date_time
            else None,
            "endDateTime": int(self._end_date_time.timestamp() * 1000.0)
            if self._end_date_time
            else None,
            "cronExpression": self._cron_expression,
            "enabled": self._enabled,
        }

    def json(self):
        return json.dumps(self, cls=util.FeatureStoreEncoder)

    @property
    def id(self):
        """Return the schedule id"""
        return self._id

    @property
    def start_date_time(self):
        """Return the schedule start time"""
        return self._start_date_time

    @property
    def end_date_time(self):
        """Return the schedule end time"""
        return self._end_date_time

    @property
    def enabled(self):
        """Return whether the schedule is enabled or not"""
        return self._enabled

    @property
    def cron_expression(self):
        """Return the schedule cron expression"""
        return self._cron_expression

    @property
    def next_execution_date_time(self):
        """Return the next execution time"""
        return self._next_execution_date_time



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json
import warnings
from typing import List, Union, Optional

from datetime import datetime, date

from hsfs import engine, statistics, util, split_statistics
from hsfs.client import exceptions
from hsfs.core import statistics_api, job
from hsfs.core.feature_descriptive_statistics import FeatureDescriptiveStatistics


class StatisticsEngine:
    def __init__(self, feature_store_id, entity_type):
        self._statistics_api = statistics_api.StatisticsApi(
            feature_store_id, entity_type
        )

    def compute_and_save_statistics(
        self,
        metadata_instance,
        feature_dataframe=None,
        feature_group_commit_id=None,
        feature_view_obj=None,
    ) -> Union[statistics.Statistics, job.Job]:
        """Compute statistics for a dataframe and send the result json to Hopsworks.
        Args:
            metadata_instance: Union[FeatureGroup, TrainingDataset]. Metadata of the entity containing the data.
            feature_dataframe: Spark or Pandas DataFrame to compute the statistics on.
            feature_group_commit_id: int. Feature group commit id.
            feature_view_obj: FeatureView. Metadata of the feature view, used when computing statistics for a Training Dataset.
        Returns:
            Union[Statistics, Job]. If running on Spark, statistics metadata containing a list of single feature descriptive statistics.
                                    Otherwise, Spark job metadata used to compute the statistics.
        """
        if engine.get_type().startswith("spark") or feature_view_obj is not None:
            # If the feature dataframe is None, then trigger a read on the metadata instance
            # We do it here to avoid making a useless request when using the Python engine
            # and calling compute_and_save_statistics
            if feature_dataframe is None:
                if feature_group_commit_id is not None:
                    feature_dataframe = (
                        metadata_instance.select_all()
                        .as_of(
                            util.get_hudi_datestr_from_timestamp(
                                feature_group_commit_id
                            )
                        )
                        .read(online=False, dataframe_type="default", read_options={})
                    )
                else:
                    feature_dataframe = metadata_instance.read()

            computation_time = int(float(datetime.now().timestamp()) * 1000)
            stats_str = self.profile_statistics_with_config(
                feature_dataframe, metadata_instance.statistics_config
            )
            desc_stats = self._parse_deequ_statistics(stats_str)
            if desc_stats:
                stats = statistics.Statistics(
                    computation_time=computation_time,
                    feature_descriptive_statistics=desc_stats,
                    window_end_commit_time=feature_group_commit_id,
                )
                return self._save_statistics(stats, metadata_instance, feature_view_obj)
        else:
            # Python engine
            return engine.get_instance().profile_by_spark(metadata_instance)

    def compute_and_save_monitoring_statistics(
        self,
        metadata_instance,
        feature_dataframe,
        window_start_commit_time,
        window_end_commit_time,
        row_percentage,
        feature_name=None,
    ) -> statistics.Statistics:
        """Compute statistics for one or more features and send the result to Hopsworks.
        Args:
            metadata_instance: Union[FeatureGroup, TrainingDataset]. Metadata of the entity containing the data.
            feature_dataframe: Spark or Pandas DataFrame to compute the statistics on.
            window_start_commit_time: int. Window start commit time
            window_end_commit_time: int. Window end commit time
            row_percentage: float. Percentage of rows to include.
            feature_name: Optional[Union[str, List[str]]]. Feature name or list of names to compute the statistics on. If not set, statistics are computed on all features.
        Returns:
            Statistics. Statistics metadata containing a list of single feature descriptive statistics.
        """
        feature_names = []
        if feature_name is None:
            feature_names = feature_dataframe.columns
        elif isinstance(feature_name, str):
            feature_names = [feature_name]
        elif isinstance(feature_name, list):
            feature_names = feature_name

        if engine.get_type() == "spark":
            commit_time = int(float(datetime.now().timestamp()) * 1000)
            stats_str = self.profile_statistics(
                feature_dataframe, feature_names, False, False, False
            )
            desc_stats = self._parse_deequ_statistics(stats_str)

            stats = statistics.Statistics(
                computation_time=commit_time,
                row_percentage=row_percentage,
                feature_descriptive_statistics=desc_stats,
                window_end_commit_time=window_end_commit_time,
                window_start_commit_time=window_start_commit_time,
            )
            return self._save_statistics(stats, metadata_instance, None)
        else:
            # TODO: Only compute statistics with Spark at the moment. This method is expected to be called
            # only through run_feature_monitoring(), which is the entrypoint of the feature monitoring job.
            # Pending work for next sprint is to compute statistics on the Python client as well, as part of
            # the deequ replacement work.
            raise exceptions.FeatureStoreException(
                "Descriptive statistics for feature monitoring cannot be computed with the Python engine."
            )

    @staticmethod
    def profile_statistics_with_config(feature_dataframe, statistics_config) -> str:
        """Compute statistics on a feature DataFrame based on a given configuration.
        Args:
            feature_dataframe: Spark or Pandas DataFrame to compute the statistics on.
            statistics_config: StatisticsConfig. Configuration for the statistics to be computed.
        Returns:
            str. Serialized features statistics.
        """
        return StatisticsEngine.profile_statistics(
            feature_dataframe,
            statistics_config.columns,
            statistics_config.correlations,
            statistics_config.histograms,
            statistics_config.exact_uniqueness,
        )

    @staticmethod
    def profile_statistics(
        feature_dataframe, columns, correlations, histograms, exact_uniqueness
    ) -> str:
        """Compute statistics on a feature DataFrame.
        Args:
            feature_dataframe: Spark or Pandas DataFrame to compute the statistics on.
            columns: List[str]. List of feature names to compute the statistics on.
            correlations: bool. Whether to compute correlations or not.
            histograms: bool. Whether to compute histograms or not.
            exact_uniqueness: bool. Whether to compute exact uniqueness values or not.
        Returns:
            str. Serialized features statistics.
        """
        if len(feature_dataframe.head(1)) == 0:
            warnings.warn(
                "There is no data in the entity that you are trying to compute "
                "statistics for. A possible cause might be that you inserted only data "
                "to the online storage of a feature group.",
                category=util.StatisticsWarning,
            )
            # if empty data, set count to 0 and return
            col_stats = [{"column": col_name, "count": 0} for col_name in columns]
            return json.dumps({"columns": col_stats})
        return engine.get_instance().profile(
            feature_dataframe, columns, correlations, histograms, exact_uniqueness
        )

    def compute_and_save_split_statistics(
        self, td_metadata_instance, feature_view_obj=None, feature_dataframes=None
    ) -> statistics.Statistics:
        """Compute statistics on Training Dataset splits

        Args:
            td_metadata_instance: TrainingDataset. Training Dataset containing the splits.
            feature_view_obj: FeatureView. Metadata of the feature view used to create the Training Dataset. This parameter is optional.
            feature_dataframes: Spark or Pandas DataFrames containing the splits to compute the statistics on.

        Returns:
            Statistics. Statistics metadata containing a list of single feature descriptive statistics.
        """
        statistics_of_splits = []
        for split in td_metadata_instance.splits:
            split_name = split.name
            stats_str = self.profile_statistics_with_config(
                (
                    feature_dataframes.get(split_name)
                    if feature_dataframes
                    else td_metadata_instance.read(split_name)
                ),
                td_metadata_instance.statistics_config,
            )
            desc_stats = self._parse_deequ_statistics(stats_str)
            statistics_of_splits.append(
                split_statistics.SplitStatistics(
                    name=split_name,
                    feature_descriptive_statistics=desc_stats,
                )
            )
        computation_time = int(float(datetime.now().timestamp()) * 1000)
        stats = statistics.Statistics(
            computation_time=computation_time, split_statistics=statistics_of_splits
        )
        return self._save_statistics(stats, td_metadata_instance, feature_view_obj)

    def compute_transformation_fn_statistics(
        self,
        td_metadata_instance,
        columns,
        label_encoder_features,
        feature_dataframe=None,
        feature_view_obj=None,
    ) -> statistics.Statistics:
        """Compute statistics for transformation functions.
        Args:
            td_metadata_instance: TrainingDataset. Training Dataset containing the splits.
            columns: List[str]. List of feature names where transformation functions are applied, excluding label encoded features.
            label_encoder_features: List[str]. List of label encoded feature names.
            feature_dataframe: Spark or Pandas DataFrame to compute the statistics on. This parameter is optional.
            feature_view_obj: FeatureView. Metadata of the feature view used to create the Training Dataset. This parameter is optional.
        Returns:
            Statistics. Statistics metadata containing a list of single feature descriptive statistics.
        """
        computation_time = int(float(datetime.now().timestamp()) * 1000)
        stats_str = self._profile_transformation_fn_statistics(
            feature_dataframe, columns, label_encoder_features
        )
        desc_stats = self._parse_deequ_statistics(stats_str)
        stats = statistics.Statistics(
            computation_time=computation_time,
            feature_descriptive_statistics=desc_stats,
            before_transformation=True,
        )
        return self._save_statistics(stats, td_metadata_instance, feature_view_obj)

    def get(
        self,
        metadata_instance,
        feature_names: Optional[List[str]] = None,
        computation_time: Optional[Union[str, int, float, datetime, date]] = None,
        before_transformation: Optional[bool] = None,
        training_dataset_version: Optional[int] = None,
    ) -> Optional[statistics.Statistics]:
        """Get statistics of an entity computed at a specific time.
           If the computation time is not provided, the most recently computed statistics will be retrieved.

        Args:
            metadata_instance: Union[FeatureGroup, TrainingDataset]. Metadata of the entity containing the data.
            feature_names: List[str]. List of feature names of which statistics are retrieved.
            computation_time: Union[str, int, float, datetime, date]. Timestamp or computation time when statistics where computed.
            before_transformation: bool. Whether the statistics were computed before transformation functions or not.
            training_dataset_version: int. Version of the training dataset on which statistics were computed.
        Returns:
            Statistics. Statistics metadata containing a list of single feature descriptive statistics.
        """
        computation_timestamp = util.convert_event_time_to_timestamp(computation_time)
        try:
            return self._statistics_api.get(
                metadata_instance,
                feature_names=feature_names,
                computation_time=computation_timestamp,
                before_transformation=before_transformation,
                training_dataset_version=training_dataset_version,
            )
        except exceptions.RestAPIError as e:
            if (
                # statistics not found
                e.response.json().get("errorCode", "")
                == exceptions.RestAPIError.FeatureStoreErrorCode.STATISTICS_NOT_FOUND
                and e.response.status_code == 404
            ):
                return None
            raise e

    def get_all(
        self,
        metadata_instance,
        feature_names: Optional[List[str]] = None,
        computation_time: Optional[Union[str, int, float, datetime, date]] = None,
        training_dataset_version: Optional[int] = None,
    ) -> Optional[List[statistics.Statistics]]:
        """Get all statistics of an entity computed before a specific time.
           If the computation time is not provided, all the statistics will be retrieved.

        Args:
            metadata_instance: Union[FeatureGroup, TrainingDataset]. Metadata of the entity containing the data.
            feature_names: List[str]. List of feature names of which statistics are retrieved.
            computation_time: Union[str, int, float, datetime, date]. Timestamp or computation time when statistics where computed.
            training_dataset_version: int. Version of the training dataset on which statistics were computed.
        Returns:
            Statistics. Statistics metadata containing a list of single feature descriptive statistics.
        """
        try:
            return self._statistics_api.get_all(
                metadata_instance,
                feature_names=feature_names,
                computation_time=computation_time,
                training_dataset_version=training_dataset_version,
            )
        except exceptions.RestAPIError as e:
            if (
                # statistics not found
                e.response.json().get("errorCode", "")
                == exceptions.RestAPIError.FeatureStoreErrorCode.STATISTICS_NOT_FOUND
                and e.response.status_code == 404
            ):
                return None
            raise e

    def get_by_time_window(
        self,
        metadata_instance,
        start_commit_time: Optional[Union[str, int, datetime, date]] = None,
        end_commit_time: Optional[Union[str, int, datetime, date]] = None,
        feature_names: Optional[List[str]] = None,
        row_percentage: Optional[float] = None,
    ) -> Union[statistics.Statistics, List[statistics.Statistics], None]:
        """Get the statistics of an entity based on a commit time window.
        Args:
            metadata_instance: Union[FeatureGroup]: Metadata of the entity containing the data.
            start_commit_time: int: Window start commit time
            end_commit_time: int: Window end commit time
            feature_names: List[str]. List of feature names of which statistics are retrieved.
            row_percentage: float. Percentage of feature values used during statistics computation
        Returns:
            Statistics:  Statistics metadata containing a list of single feature descriptive statistics.
        """
        start_commit_time = util.convert_event_time_to_timestamp(start_commit_time)
        end_commit_time = util.convert_event_time_to_timestamp(end_commit_time)
        try:
            return self._statistics_api.get(
                metadata_instance,
                start_commit_time=start_commit_time,
                end_commit_time=end_commit_time,
                feature_names=feature_names,
                row_percentage=row_percentage,
            )
        except exceptions.RestAPIError as e:
            if (
                # statistics not found
                e.response.json().get("errorCode", "")
                == exceptions.RestAPIError.FeatureStoreErrorCode.STATISTICS_NOT_FOUND
                and e.response.status_code == 404
            ) or (
                # commits not found
                e.response.json().get("errorCode", "")
                == exceptions.RestAPIError.FeatureStoreErrorCode.FEATURE_GROUP_COMMIT_NOT_FOUND
                and e.response.status_code == 400
            ):
                return None
            raise e

    def _profile_transformation_fn_statistics(
        self, feature_dataframe, columns, label_encoder_features
    ) -> str:
        if (
            engine.get_type() == "spark"
            and len(feature_dataframe.select(*columns).head(1)) == 0
        ) or (
            (engine.get_type() == "hive" or engine.get_type() == "python")
            and len(feature_dataframe.head()) == 0
        ):
            raise exceptions.FeatureStoreException(
                "There is no data in the entity that you are trying to compute "
                "statistics for. A possible cause might be that you inserted only data "
                "to the online storage of a feature group."
            )

        # compute statistics for all features with transformation fn
        all_columns = (columns or []) + (label_encoder_features or [])
        stats_str = engine.get_instance().profile(
            feature_dataframe, all_columns, False, True, False
        )

        # add unique values profile to column stats
        return self._profile_unique_values(
            feature_dataframe, label_encoder_features, stats_str
        )

    def _profile_unique_values(
        self, feature_dataframe, label_encoder_features, stats_str
    ) -> str:
        stats = json.loads(stats_str)
        if not stats:
            stats = {"columns": []}
        stats_dict = {col_stats["column"]: col_stats for col_stats in stats["columns"]}
        for column in label_encoder_features:
            col_stats_unique_values = {
                "column": column,
                "unique_values": [
                    value
                    for value in engine.get_instance().get_unique_values(
                        feature_dataframe, column
                    )
                ],
            }
            if column in stats_dict:
                stats_dict[column].update(col_stats_unique_values)
            else:
                stats_dict[column] = col_stats_unique_values

        stats["columns"] = list(stats_dict.values())
        return json.dumps(stats)  # the result is a JSON string

    def _save_statistics(
        self, stats, metadata_instance, feature_view_obj
    ) -> statistics.Statistics:
        # metadata_instance can be feature group or training dataset
        if feature_view_obj:
            stats = self._statistics_api.post(
                feature_view_obj,
                stats=stats,
                training_dataset_version=metadata_instance.version,
            )
        else:
            stats = self._statistics_api.post(
                metadata_instance, stats=stats, training_dataset_version=None
            )
        return stats

    def _parse_deequ_statistics(self, stats) -> List[FeatureDescriptiveStatistics]:
        if stats is None:
            warnings.warn(
                "There is no Deequ statistics to deserialize. A possible cause might be that Deequ did not succeed in the statistics computation.",
                category=util.StatisticsWarning,
            )
            return None
        if isinstance(stats, str):
            stats = json.loads(stats)
        return [
            FeatureDescriptiveStatistics.from_deequ_json(col_stats)
            for col_stats in stats["columns"]
        ]



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from typing import List, Union
from hsfs.core import validation_report_api
from hsfs import client, util
import great_expectations as ge

from hsfs.validation_report import ValidationReport


class ValidationReportEngine:
    def __init__(self, feature_store_id: int, feature_group_id: int):
        """Validation Report engine.

        :param feature_store_id: id of the respective featurestore
        :type feature_store_id: int
        :param feature_group_id: id of the featuregroup it is attached to
        :type feature_group_id: int
        """
        self._feature_store_id = feature_store_id
        self._feature_group_id = feature_group_id
        self._validation_report_api = validation_report_api.ValidationReportApi(
            feature_store_id=feature_store_id, feature_group_id=feature_group_id
        )

    def save(
        self, validation_report: ValidationReport, ge_type: bool = True
    ) -> ValidationReport:
        saved_report = self._validation_report_api.create(validation_report)
        url = self._get_validation_report_url()
        print(f"Validation Report saved successfully, explore a summary at {url}")
        if ge_type:
            return saved_report.to_ge_type()
        else:
            return saved_report

    def get_last(
        self, ge_type: bool = True
    ) -> Union[ValidationReport, ge.core.ExpectationSuiteValidationResult, None]:
        """Get the most recent Validation Report of a Feature Group."""
        url = self._get_validation_report_url()
        print(
            f"""Long reports can be truncated when fetching from Hopsworks.
        \nYou can download the full report at {url}"""
        )
        reports = self._validation_report_api.get_last()
        if len(reports) == 0:
            return None
        elif len(reports) == 1 and ge_type:
            return reports[0].to_ge_type()
        else:
            return reports[0]

    def get_all(
        self, ge_type: bool = True
    ) -> Union[List[ValidationReport], List[ge.core.ExpectationSuiteValidationResult]]:
        """Get all Validation Report of a Feature Group."""
        url = self._get_validation_report_url()
        print(
            f"""Long reports can be truncated when fetching from Hopsworks.
        \nYou can download full reports at {url}"""
        )
        reports = self._validation_report_api.get_all()
        if ge_type:
            return [report.to_ge_type() for report in reports]
        else:
            return reports

    def delete(self, validation_report_id: int):
        self._validation_report_api.delete(validation_report_id)

    def _get_validation_report_url(self) -> str:
        """Build url to land on Hopsworks UI page which summarizes validation results"""
        sub_path = (
            "/p/"
            + str(client.get_instance()._project_id)
            + "/fs/"
            + str(self._feature_store_id)
            + "/fg/"
            + str(self._feature_group_id)
        )
        return util.get_hostname_replaced_url(sub_path)
